from plotly.colors import hex_to_rgb, n_colors


def chunks(input_list, number_of_chunks):
    """Yield number_of_chunks number of striped chunks from input_list."""
    for i in range(0, number_of_chunks):
        yield input_list[i::number_of_chunks]


def rgb_to_hex(rgb: tuple = None) -> str:
    """Convert RGB to hex

    :param rgb: tuple in format (r, g, b) where r, g, b are integers in range [0-255]
    :return str: converted color as hex-string
    """
    # necessary due to weird behaviour introduced by plotly
    # test = plotly.colors.ncolors(lowcolor=(64, 60, 83), highcolor=(255, 0, 255), n_colors=11001)
    # print(test[-1])
    if any(False if 0 <= _ < 256 else True for _ in rgb):
        for idx, color in enumerate(rgb):
            if color < 0:
                tmp = list(rgb)
                tmp[idx] = 0
                print(f"Corrected color: {rgb} > {tuple(list(tmp))}")
                rgb = tuple(list(tmp))
    return f"#{int(rgb[0]):02X}{int(rgb[1]):02X}{int(rgb[2]):02X}"


def generate_color_range(start_color: str = None, stop_color: str = None,
                         values: int = None) -> list:
    """Calculate color range based on start and stop color and number of values

    :param start_color: start color as RGB hex
    :param stop_color: stop color as RGB hex
    :param values: number of colors to generate
    :return list: list of hex color codes
    """
    start_color = hex_to_rgb(start_color)
    stop_color = hex_to_rgb(stop_color)

    try:
        color_list = n_colors(start_color, stop_color, values)
        return [rgb_to_hex(_) for _ in color_list]
    except ZeroDivisionError:
        return [rgb_to_hex(start_color)]


def get_brightness(rgb_color: tuple = None) -> float:
    """
    Get the brightness of a rgb color triplet

    :param rgb_color: rgb color triplet
    :return: brightness as float
    """
    r, g, b = rgb_color
    brightness = (r * 299 + g * 587 + b * 114) / 1000
    return brightness


def prioritize_bright_colors(color_list: list = None) -> list:
    """
    Filters a list of threshold-color-tuples (0.0, "#FFFFFF") and eliminates duplicates while
    prioritizing brighter colors over darker ones

    :param color_list: color list in the format [(0.0, "#585569"), (0.0, "#FEFEFE"), (0.1, ..), ..]
    :return: filtered color list in the format [(0.0, "#FEFEFE"), (0.1, ..), ..]
    """
    sorted_colors = {}
    for idx, color in color_list:
        brightness = get_brightness(hex_to_rgb(color))
        if idx not in sorted_colors or get_brightness(hex_to_rgb(sorted_colors[idx])) < brightness:
            sorted_colors[idx] = color

    return [(idx, color) for idx, color in sorted_colors.items()]



import os
import tarfile
import json
from functools import partial
from threading import Thread
import time
import textwrap

from tkinter import Tk, StringVar, BooleanVar, IntVar, filedialog, messagebox, END
from tkinter.ttk import LabelFrame, Frame, Style

from src.ontoloviz.core import MeSHSunburst, ATCSunburst
from src.ontoloviz.obo_utils import get_remote_ontology, build_non_separator_based_tree
from src.ontoloviz.app_utils import Button, Entry, Combobox, Checkbutton, Label, Radiobutton
from src.ontoloviz.app_utils import create_tooltip, exception_as_popup
from src.ontoloviz.app_utils import ExportPopup, ColorScalePopup, BorderPopup, SelectOptionsPopup


class App(Tk):
    """OntoloViz App class"""
    def __init__(self):
        """App Initialization, styles, memory variables"""
        super().__init__()
        self.title("OntoloViz")
        self.resizable(True, True)
        self.minsize(400, 80)

        # highres settings for screenshot
        # import ctypes
        # ctypes.windll.shcore.SetProcessDpiAwareness(1)
        # self.tk.call("tk", "scaling", 2)

        # style definitions
        self.d4_red = "#C33D35"
        self.d4_purple = "#403C53"
        self.d4_custom = "#8CA6D9"
        self.d4_green = "#579D66"
        self.d4_white = "#FFFFFF"
        self.d4_black = "#000000"
        dark_bg = "#2E2D32"
        success = "#6BBE92"
        self.bold_normal = ("Arial", 8, "bold")
        self.bold_large = ("Arial", 9, "bold")
        self.configure(background=dark_bg)
        self.style = Style()
        self.style.configure("success.TButton", font=self.bold_large, background=success)
        self.style.configure("dark.TButton", font=self.bold_normal, background=dark_bg)
        self.style.configure("dark.TFrame", background=dark_bg)
        self.style.configure("dark.TLabelframe", font=self.bold_large, background=dark_bg,
                             relief="ridge")
        self.style.configure("dark.TLabelframe.Label", background=dark_bg, foreground=self.d4_white)
        self.style.configure("dark.TLabel", background=dark_bg, foreground=self.d4_white)
        self.change_theme_color(foreground=self.d4_white, background=self.d4_red)

        # core variables
        self.p = MeSHSunburst()
        self.d = ATCSunburst()
        self.obo = None

        # memory variables (settings)
        self.database_var = StringVar()
        self.color_scale_var = StringVar(
            value=f'[[0, "{self.d4_white}"], [0.2, "{self.d4_purple}"], [1, "{self.d4_red}"]]')
        self.show_border_var = BooleanVar(value=True)
        self.border_color = StringVar(value="rgba(0,0,0,0.25)")
        self.border_width = StringVar(value="1")
        self.export_plot_var = BooleanVar(value=False)
        self.mesh_data_sources = ["Utilization Tuple: Semantic Direct",
                                  "Utilization Tuple: Semantic Indirect",
                                  "Utilization Tuple: Explicit Direct",
                                  "Utilization Tuple: Explicit Indirect"]
        self.mesh_data_source_var = StringVar(value=self.mesh_data_sources[0])
        self.mesh_drop_empty_var = BooleanVar(value=False)
        self.mesh_legend_enabled_control = BooleanVar(value=True)
        self.mesh_legend_enable = None  # Checkbutton
        self.mesh_propagate_enabled_control = BooleanVar(value=False)
        self.mesh_propagate_enable = None  # Checkbutton
        self.mesh_propagate_lvl_var = IntVar(value=0)
        self.mesh_propagate_lvl_lbl = None  # label 'Level: '
        self.mesh_propagate_lvl = None  # Combobox
        self.mesh_propagate_color_var = StringVar(value="specific")
        self.mesh_propagate_color = None  # Combobox
        self.mesh_propagate_color_lbl = None
        self.mesh_propagate_counts_var = StringVar(value="off")
        self.mesh_propagate_counts = None  # Combobox
        self.mesh_propagate_counts_lbl = None
        self.mesh_label_var = StringVar(value="all")
        self.mesh_asset_var = StringVar()
        self.mesh_summary_plot_var = IntVar(value=5)  # stores values from entry
        self.mesh_summary_plot_control = BooleanVar(value=True)  # controls enabling of entry
        self.mesh_summary_plot = None  # checkbox widget
        self.mesh_summary_plot_cols = None  # entry widget
        self.mesh_summary_plot_lbl = None  # label 'Columns: '
        self.mesh_plot_type_var = StringVar(value="Sunburst Plot")
        self.mesh_plot_type = None  # Combobox
        self.atc_data_sources = ["Linked Tuple"]
        self.atc_data_source_var = StringVar(value=self.atc_data_sources[0])
        self.atc_legend_enabled_control = BooleanVar(value=True)
        self.atc_legend_enable = None  # Checkbutton
        self.atc_propagate_enabled_control = BooleanVar(value=False)
        self.atc_propagate_enable = None  # Checkbutton
        self.atc_propagate_lvl_var = IntVar(value=5)
        self.atc_propagate_lvl_lbl = None  # label 'Level: '
        self.atc_propagate_lvl = None  # Combobox
        self.atc_propagate_color_var = StringVar(value="specific")
        self.atc_propagate_color = None  # Combobox
        self.atc_propagate_color_lbl = None
        self.atc_propagate_counts_var = StringVar(value="off")
        self.atc_propagate_counts = None  # Combobox
        self.atc_propagate_counts_lbl = None
        self.atc_label_var = StringVar(value="all")
        self.atc_wedge_width_var = StringVar(value="total")
        self.atc_asset_var = StringVar()
        self.atc_summary_plot_var = IntVar(value=5)
        self.atc_summary_plot_control = BooleanVar(value=True)
        self.atc_summary_plot = None
        self.atc_summary_plot_cols = None
        self.atc_summary_plot_lbl = None
        self.loaded_settings = {}
        self.status_var = StringVar()
        self.mesh_frame = None
        self.atc_frame = None
        self.status_frame = None
        self.recent_ui_toggle_mode = None
        self.color_scale = None
        self.color_scale_btn_mesh = None
        self.color_scale_btn_atc = None
        self.color_scale_tt_template = None
        self.show_border = None
        self.show_border_btn_mesh = None
        self.show_border_btn_atc = None
        self.show_border_tt_template = None
        self.load_file_btn = None
        self.load_obo_url_btn = None
        self.atc_file_loaded = ""
        self.mesh_file_loaded = ""
        self.performance_warning_shown = False
        self.custom_ontology_separator = None

        # various templates
        self.alt_text_db = "This functionality requires a valid database"
        self.alt_text = "This functionality requires a valid database or a loaded file"
        self.color_scale_tt_template = str("Define a custom color scale for the sunburst.\n"
                                           "Requires active propagation, overwrites colors defined "
                                           "in file. \nCurrent scale: ")
        self.show_border_tt_template = "Configure the border drawn around the sunburst wedges"
        self.save_plot_tt_template = str("Save the generated plot as interactive .html file and "
                                         "creates a .tsv template for usage with own data based "
                                         "on current settings")
        self.plot_tt_template = "Generate plot and open interactive sunburst in browser"
        self.export_tt_template = str("Generate sunburst data without plotting, export to "
                                      "Excel/TSV for later use/customization")
        self.propagate_color_tt = str(
            "off: Color scale is based on 'Color' column from imported file\n"
            "specific: Color scale is based on the maximum values of the corresponding tree\n"
            "global: Color scale is based on the maximum values of the entire ontology"
            "ALT:Enable 'Propagation' to modify propagation color specificity"
        )
        self.propagate_color_mesh_tt = str(
            "off: Color scale is based on 'Color' column from imported file\n"
            "specific: Color scale is based on the max values of the corresponding tree\n"
            "global: Color scale is based on the max values of the entire ontology\n"
            "phenotype: Only the most outer node in a branch is colored"
            "ALT:Enable 'Propagation' to modify propagation color specificity"
        )
        self.propagate_counts_tt = str(
            "off: no counts are propagated, counts equal imported values\n"
            "level: counts are propagated up to defined level, values above threshold remain "
            "unchanged\n"
            "all: counts are propagated up to central node, imported values are corrected "
            "and overwritten\n"
            "ALT: Enable 'Propagation' to modify propagation counts"
        )
        self.hpo_ontology_tt = str(
            "Fetches the Human Phenotype Ontology (sub-tree 'Phenotypic "
            "abnormality') from https://purl.obolibrary.org/obo/hp.obo\n---\n"
            "The Human Phenotype Ontology (HPO) aims to provide a standardized "
            "vocabulary of phenotypic abnormalities encountered in human disease."
            "\nEach term in the HPO describes a phenotypic abnormality, such as "
            "atrial septal defect.\nThe HPO is currently being developed using the"
            " medical literature, Orphanet, DECIPHER, and OMIM."
            "\n---\nFor more information visit: https://hpo.jax.org/app"
        )
        self.gene_ontology_tt = str(
            "(sub-trees with < 2 nodes excluded) from "
            "https://current.geneontology.org/ontology/go.obo\n---\n"
            "The goal of the GeneOntology (GO) project is to provide a uniform "
            "way to describe the functions of gene products\nfrom organisms "
            "across all kingdoms of life and thereby enable analysis of genomic "
            "data\n---\nFor more information visit: http://geneontology.org/"
        )
        self.po_ontology_tt = str(
            "Fetches the Plant Ontology (sub-tree 'plant structure') "
            "(sub-trees with < 5 nodes excluded) "
            "from https://purl.obolibrary.org/obo/po.obo\n---\n"
            "The Plant Ontology is a structured vocabulary and database resource that links plant "
            "anatomy, morphology and growth and development to plant genomics data.\nThe PO is "
            "under active development to expand to encompass terms and annotations from all plants."
            "\n---\nFor more information visit: http://planteome.org/"
        )
        self.cl_ontology_tt = str(
            "Fetches the Cell Ontology (without inter-ontology axioms) "
            "(sub-trees with < 2 nodes excluded) "
            "from http://purl.obolibrary.org/obo/cl/cl-basic.obo\n---\n"
            "The Cell Ontology is designed as a structured controlled vocabulary for cell types.\n"
            "This ontology was constructed for use by the model organism and other bioinformatics "
            "databases, where there is a need for a controlled vocabulary of cell types.\nThis "
            "ontology is not organism specific. It covers cell types from prokaryotes to mammals.\n"
            "However, it excludes plant cell types, which are covered by PO."
            "\n---\nFor more information visit: https://biccn.org/"
        )
        self.chebi_ontology_tt = str(
            "Fetches the Chemical Entities of Biological Interest (CHEBI) Ontology (sub-tree "
            "'molecular entity') from https://purl.obolibrary.org/obo/chebi/chebi_lite.obo\n---\n"
            "A freely available dictionary of molecular entities focused on ‘small’ chemical "
            "compounds.\nThe term ‘molecular entity’ refers to any constitutionally or isotopically"
            " distinct atom, molecule, ion, ion pair, radical, radical ion, complex, conformer, "
            "etc., identifiable as a separately distinguishable entity.\nThe molecular entities in "
            "question are either products of nature or synthetic products used to intervene in the "
            "processes of living organisms."
            "\n---\nFor more information visit: https://www.ebi.ac.uk/chebi/"
        )
        self.uberon_ontology_tt = str(
            "Fetches the Uberon multi-species anatomy ontology (sub-tree 'anatomical structure') "
            "(sub-trees with < 2 nodes excluded) from "
            "https://purl.obolibrary.org/obo/uberon/basic.obo\n---\n"
            "Uberon is an integrated cross-species ontology covering anatomical structures in "
            "animals.\n---\nFor more information visit: https://obophenotype.github.io/uberon/"
        )
        self.doid_ontology_tt = str(
            "Fetches the Human Disease Ontology from https://purl.obolibrary.org/obo/doid.obo"
            "\n---\nThe Disease Ontology has been developed as a standardized ontology for human "
            "disease with the purpose of providing the biomedical community with consistent, \n"
            "reusable and sustainable descriptions of human disease terms, phenotype "
            "characteristics and related medical vocabulary disease concepts through "
            "collaborative efforts with biomedical researchers, \ncoordinated by the University "
            "of Maryland School of Medicine, Institute for Genome Sciences.\nThe Disease Ontology "
            "semantically integrates disease and medical vocabularies through extensive cross "
            "mapping of DO terms to MeSH, ICD, NCI’s thesaurus, SNOMED and OMIM."
            "\n---\nFor more information visit: https://www.disease-ontology.org/"
        )

        # function calls
        self.build_base_ui()
        self.toggle_widgets(enable=False, mode="db")

    def change_theme_color(self, foreground: str = None, background: str = None) -> None:
        self.style.configure("primary.TLabelframe", background=background, relief="ridge")
        self.style.configure("primary.TLabelframe.Label", font=self.bold_large,
                             background=background, foreground=foreground)
        self.style.configure("primary_sub.TLabelframe", background=background)
        self.style.configure("primary_sub.TLabelframe.Label", background=background,
                             foreground=foreground, font=self.bold_normal)
        self.style.configure("primary.TFrame", background=background)
        self.style.configure("primary.TLabel", background=background, foreground=foreground)
        self.style.configure("primary.TButton", font=self.bold_normal, background=background)
        self.style.configure("primary.TCheckbutton", background=background,
                             activebackground=background, foreground=foreground)
        self.style.configure("primary.TRadiobutton", background=background,
                             activebackground=background)

    def build_base_ui(self):
        """Builds the base graphical UI elements to load a file"""

        # ######################################### LOAD FRAME AT TOP ############################ #

        load_frm = Frame(self, style="dark.TFrame")
        load_frm.pack(ipadx=2, ipady=2, fill="both", expand=True)

        self.load_file_btn = Button(load_frm, text="Load File", command=self.load_file,
                                    style="success.TButton")
        self.load_file_btn.pack(side="left", padx=(4, 2), pady=(2, 0))
        create_tooltip(self.load_file_btn,
                       "Load an Excel or .tsv file with ontology-based data."
                       "\n ---"
                       "\n - ATC/MeSH ontologies are automatically recognized if "
                       "given templates are used"
                       "\n - For other ontologies, the separator has to be defined after "
                       "loading the file (possible separators: . , _ /)"
                       "\n ---"
                       "\n - Minimal Example for separator-based ontologies "
                       "(requires 5 columns and header):"
                       "\n - parents are automatically created if not defined"
                       "\n   -------------------------------------------------------"
                       "\n   ID              | Label   | Description | Count | Color"
                       "\n   --              | -----   | ----------- | ----- | -----"
                       "\n   A               | group 1 |             |       |"
                       "\n   A_1             | child 1 |             |       |"
                       "\n   B_1_2_3|C_1_2_3 | child 2 |             |       |"
                       "\n   -------------------------------------------------------"
                       "\n - Minimal Example for non-structured ontologies "
                       "(requires 6 columns and header):"
                       "\n - nodes without valid parent IDs are removed"
                       "\n   ----------------------------------------------------------------"
                       "\n   ID              | Parent | Label   | Description | Count | Color"
                       "\n   --              | ------ | -----   | ----------- | ----- | -----"
                       "\n   HP:00A          |        | group 1 |             |       |"
                       "\n   HP:001          | HP:00A | child 1 |             |       |"
                       "\n   HP:002          | HP:00A | child 2 |             |       |"
                       "\n   ----------------------------------------------------------------"
                       "\n ---"
                       "\n - ATC: counts for non-drug levels (1-4) will be recalculated and "
                       "overwritten if a parents value does not match all child values."
                       "\n - ATC: custom colors are not applied when propagation is active"
                       "\n ---"
                       "\n - Excel files remember settings, .tsv files always load defaults"
                       "\n - For more information, read: https://github.com/Delta4AI/OntoloViz")

        self.load_obo_url_btn = Button(load_frm, text="Load from Web", command=self.load_url,
                                       style="success.TButton")
        self.load_obo_url_btn.pack(side="left", padx=2, pady=(2, 0))
        create_tooltip(self.load_obo_url_btn, "Download and visualize .obo ontologies from a "
                                              "predefined list or specify a custom URL")

        # ####################################### MESH/DRUG FRAMES ############################### #

        self.mesh_frame = Frame(self, style="dark.TFrame")
        self.mesh_frame.pack(fill="both")
        self.atc_frame = Frame(self, style="dark.TFrame")
        self.atc_frame.pack(fill="both")

        # ###################################### STATUS AT BOTTOM ################################ #

        self.status_frame = LabelFrame(self, text="Status", style="dark.TLabelframe")
        self.status_frame.pack(ipadx=2, ipady=2, fill="both")
        status = Label(self.status_frame, textvariable=self.status_var, style="dark.TLabel")
        status.pack(side="right", padx=2)

    def build_mesh_ui(self, db_functions: bool = None):
        """Builds MeSH-phenotype specific controls if proper file was loaded

        :param db_functions: if True, database related widgets are generated
        """
        # ###################################### PHENOTYPE/MESH SUNBURST ######################### #

        # top frame
        p_frm = LabelFrame(self.mesh_frame, text="Display Settings", style="primary.TLabelframe")
        p_frm.pack(ipadx=2, ipady=2, fill="both")

        if db_functions:
            # query data frame
            mesh_data_frm = LabelFrame(p_frm, text="Query Data", style="primary_sub.TLabelframe")
            mesh_data_frm.pack(ipadx=2, ipady=2, padx=4, pady=2, anchor="w")

            # asset subframe
            mesh_asset_frm = Frame(mesh_data_frm, style="primary.TFrame")
            mesh_asset_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")
            mesh_asset_label = Label(mesh_asset_frm, text="Asset name:", style="primary.TLabel",
                                     db_w=True, width=12)
            mesh_asset_label.pack(side="left", padx=2)
            mesh_asset = Entry(mesh_asset_frm, db_w=True, textvariable=self.mesh_asset_var,
                               width=40)
            mesh_asset.pack(side="left", padx=(2, 10))
            create_tooltip(mesh_asset, "Enter drug name name (e.g. ASPIRIN, case sensitive)")

            # data source subframe
            mesh_data_source_frm = Frame(mesh_data_frm, style="primary.TFrame")
            mesh_data_source_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")
            mesh_data_source_label = Label(mesh_data_source_frm, text="Data source:",
                                           style="primary.TLabel", db_w=True, width=12)
            mesh_data_source_label.pack(side="left", padx=2)
            mesh_data_source = Combobox(mesh_data_source_frm,
                                        textvariable=self.mesh_data_source_var,
                                        values=self.mesh_data_sources, state="readonly", db_w=True,
                                        width=max([len(_) for _ in self.mesh_data_sources]))
            mesh_data_source.pack(side="left", padx=2, fill="x", expand=True)
            create_tooltip(mesh_data_source, "Select data source")

        # frame for options
        p_options_frm = Frame(p_frm, style="primary.TFrame")
        p_options_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")

        # general options subframe
        mesh_display_options_frm = LabelFrame(p_options_frm, text="General",
                                              style="primary_sub.TLabelframe")
        mesh_display_options_frm.pack(fill="both")

        # colors subframe
        mesh_display_options_top_frm = Frame(mesh_display_options_frm, style="primary.TFrame")
        mesh_display_options_top_frm.pack(fill="both", pady=2)

        # set color scale button
        self.color_scale_btn_mesh = Button(mesh_display_options_top_frm, text="Set Color Scale",
                                           style="primary.TButton", db_w=True, mesh_w=True,
                                           command=lambda: ColorScalePopup(self))
        self.color_scale_btn_mesh.pack(side="left", padx=2)
        create_tooltip(self.color_scale_btn_mesh,
                       self.color_scale_tt_template + self.color_scale_var.get())

        # set border button
        self.show_border_btn_mesh = Button(mesh_display_options_top_frm, text="Set Border",
                                           style="primary.TButton", db_w=True, mesh_w=True,
                                           command=lambda: BorderPopup(self))
        self.show_border_btn_mesh.pack(side="left", padx=2)
        create_tooltip(self.show_border_btn_mesh, self.show_border_tt_template
                       + "\nCurrent properties: Color: " + self.border_color.get()
                       + ", Width: " + self.border_width.get())

        # display legend checkmark
        self.mesh_legend_enable = Checkbutton(mesh_display_options_top_frm, text="Legend",
                                              style="primary.TCheckbutton", db_w=True,
                                              mesh_w=True, onvalue=True, offvalue=False,
                                              variable=self.mesh_legend_enabled_control)
        self.mesh_legend_enable.pack(side="right", padx=2)
        create_tooltip(self.mesh_legend_enable,
                       "Displays a legend in form of a weighted color bar. "
                       "Disabled for summary plots with specific color propagation enabled.")

        mesh_display_options_bottom_frm = Frame(mesh_display_options_frm, style="primary.TFrame")
        mesh_display_options_bottom_frm.pack(fill="both", pady=(0, 2))

        # drop empty last child
        mesh_drop_empty = Checkbutton(mesh_display_options_bottom_frm, text="Drop empty nodes",
                                      variable=self.mesh_drop_empty_var, onvalue=True,
                                      offvalue=False, style="primary.TCheckbutton", db_w=True,
                                      mesh_w=True)
        mesh_drop_empty.pack(side="left", padx=2)
        create_tooltip(mesh_drop_empty, "Drop nodes who have no further children and 0 counts")

        # labels ('all', 'propagate', 'none')
        mesh_label = Combobox(mesh_display_options_bottom_frm, textvariable=self.mesh_label_var,
                              state="readonly", width=11, values=["all", "propagation", "none"],
                              db_w=True, mesh_w=True)
        mesh_label.pack(side="right", padx=2)
        create_tooltip(mesh_label, "Enables/Disables display of labels inside sunburst wedges")

        mesh_label_label = Label(mesh_display_options_bottom_frm, text="Display Labels:",
                                 style="primary.TLabel", db_w=True, mesh_w=True)
        mesh_label_label.pack(side="right", padx=2)

        # propagation subframe
        mesh_propagate_frm = LabelFrame(p_options_frm, text="Propagation",
                                        style="primary_sub.TLabelframe")
        mesh_propagate_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="both")
        self.mesh_propagate_enabled_control.set(False)
        self.mesh_propagate_enable = Checkbutton(mesh_propagate_frm, text="Enable",
                                                 style="primary.TCheckbutton",
                                                 db_w=True, mesh_w=True,
                                                 onvalue=True, offvalue=False,
                                                 variable=self.mesh_propagate_enabled_control,
                                                 command=partial(self.checkbox_controller,
                                                                 "mesh_propagate"))
        self.mesh_propagate_enable.pack(side="left", padx=2)
        create_tooltip(self.mesh_propagate_enable,
                       "Select to enable propagation of counts and colors based on ontology level")

        # propagate color specificity
        self.mesh_propagate_color_lbl = Label(mesh_propagate_frm, text="Color:",
                                              style="primary.TLabel")
        self.mesh_propagate_color_lbl.pack(side="left", padx=2)
        self.mesh_propagate_color_lbl.configure(state="disabled")
        create_tooltip(self.mesh_propagate_color_lbl,self.propagate_color_mesh_tt)
        self.mesh_propagate_color = Combobox(mesh_propagate_frm,
                                             textvariable=self.mesh_propagate_color_var,
                                             width=10,
                                             state="readonly",
                                             values=["off", "specific", "global", "phenotype"])
        self.mesh_propagate_color.pack(side="left", padx=2)
        self.mesh_propagate_color.configure(state="disabled")
        create_tooltip(self.mesh_propagate_color, self.propagate_color_mesh_tt)

        # propagate counts
        self.mesh_propagate_counts_lbl = Label(mesh_propagate_frm, text="Counts:",
                                               style="primary.TLabel")
        self.mesh_propagate_counts_lbl.pack(side="left", padx=2)
        self.mesh_propagate_counts_lbl.configure(state="disabled")
        create_tooltip(self.mesh_propagate_counts_lbl, self.propagate_counts_tt)
        self.mesh_propagate_counts = Combobox(mesh_propagate_frm,
                                              textvariable=self.mesh_propagate_counts_var,
                                              state="readonly",
                                              values=["off", "level", "all"],
                                              width=4)
        self.mesh_propagate_counts.pack(side="left", padx=2)
        self.mesh_propagate_counts.configure(state="disabled")
        create_tooltip(self.mesh_propagate_counts, self.propagate_counts_tt)

        # propagate level
        self.mesh_propagate_lvl_lbl = Label(mesh_propagate_frm, text="Level: ",
                                            style="primary.TLabel")
        self.mesh_propagate_lvl_lbl.pack(side="left", padx=2)
        self.mesh_propagate_lvl_lbl.configure(state="disabled")
        self.mesh_propagate_lvl = Combobox(mesh_propagate_frm,
                                           textvariable=self.mesh_propagate_lvl_var,
                                           width=3,
                                           state="readonly",
                                           values=[str(_) for _ in range(0, 14, 1)])
        self.mesh_propagate_lvl.pack(side="left", padx=2)
        self.mesh_propagate_lvl.configure(state="disabled")
        create_tooltip(self.mesh_propagate_lvl, "Propagate from outer to inner levels up to "
                                                "defined level\n"
                                                "0 corresponds to the central node, 13 to the "
                                                "outermost node\n"
                                                "---\n"
                                                "affects color propagation when set to "
                                                "'specific' or 'global'\n"
                                                "affects count propagation when set to 'level'"
                                                "ALT:Enable 'Propagation' to modify "
                                                "propagation level")

        # summary plot subframe
        mesh_summary_plot_frm = LabelFrame(p_options_frm, text="Plot",
                                           style="primary_sub.TLabelframe")
        mesh_summary_plot_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="both")

        # checkbutton to toggle overview / detailed view
        self.mesh_summary_plot_cols = Entry(mesh_summary_plot_frm, width=2, validate="focusout",
                                            validatecommand=partial(self.overview_entry_validation,
                                                                    "mesh"))
        self.mesh_summary_plot_cols.insert(0, "5")
        self.mesh_summary_plot = Checkbutton(mesh_summary_plot_frm, text="Summary",
                                             style="primary.TCheckbutton", db_w=True, mesh_w=True,
                                             variable=self.mesh_summary_plot_control, onvalue=True,
                                             offvalue=False,
                                             command=partial(self.checkbox_controller,
                                                             "mesh_summary_plot"))
        self.mesh_summary_plot.pack(side="left", padx=2)
        self.mesh_summary_plot_lbl = Label(mesh_summary_plot_frm, text="Columns: ",
                                           style="primary.TLabel")
        self.mesh_summary_plot_lbl.pack(side="left", padx=2)
        self.mesh_summary_plot_cols.pack(side="left", padx=2)
        create_tooltip(self.mesh_summary_plot_cols,
                       "Enter amount of columns in range (1..20)ALT:Enable 'Summary Plot' "
                       "to modify amount of columns")
        create_tooltip(self.mesh_summary_plot,
                       "Select to plot all data in a combined overview (resource intensive, "
                       "set Labels to 'none' for faster loading)")

        self.mesh_plot_type = Combobox(mesh_summary_plot_frm,
                                       textvariable=self.mesh_plot_type_var,
                                       state="readonly",
                                       values=[str(_) for _ in self.p.plot_type.keys()],
                                       width=14,
                                       mesh_w=True)
        self.mesh_plot_type.pack(side="right", padx=2)
        create_tooltip(self.mesh_plot_type, "Select type of plot to generate")

        # run buttons frame
        p_run_frm = Frame(p_frm, style="primary.TFrame")
        p_run_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="x")

        # plot button
        mesh_plot = Button(p_run_frm, text="Plot", style="primary.TButton",
                           command=partial(self.plot, "mesh"), db_w=True, mesh_w=True)
        mesh_plot.pack(side="right", padx=2)
        create_tooltip(mesh_plot, self.plot_tt_template)

        # save plot button
        save_plot_btn = Checkbutton(p_run_frm, text="Save", style="primary.TCheckbutton",
                                    variable=self.export_plot_var, db_w=True, mesh_w=True)
        save_plot_btn.pack(side="right", padx=2)
        create_tooltip(save_plot_btn, self.save_plot_tt_template)

        if db_functions:
            # export button
            mesh_export = Button(p_run_frm, text="Export", style="primary.TButton",
                                 command=self.mesh_export, db_w=True)
            mesh_export.pack(side="right", padx=2)
            create_tooltip(mesh_export, self.export_tt_template)

    def build_atc_ui(self, db_functions: bool = None):
        """Builds ATC-drug specific controls if proper file was loaded

        :param db_functions: if True, database related widgets are generated
        """

        # ###################################### DRUG/ATC SUNBURST ############################### #

        # top frame
        d_frm = LabelFrame(self.atc_frame, text="Display Settings", style="primary.TLabelframe")
        d_frm.pack(ipadx=2, ipady=2, fill="both")

        if db_functions:
            # query data frame
            atc_data_frm = LabelFrame(d_frm, text="Query Data", style="primary_sub.TLabelframe")
            atc_data_frm.pack(ipadx=2, ipady=2, padx=4, pady=2, anchor="w")

            # asset subframe
            atc_asset_frm = Frame(atc_data_frm, style="primary.TFrame")
            atc_asset_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")
            atc_asset_label = Label(atc_asset_frm, text="Asset name:", style="primary.TLabel",
                                    db_w=True, width=12)
            atc_asset_label.pack(side="left", padx=2)
            atc_asset = Entry(atc_asset_frm, db_w=True, textvariable=self.atc_asset_var, width=40)
            atc_asset.pack(side="left", padx=(2, 10))
            create_tooltip(atc_asset, "Enter phenotype name (e.g. headache, case sensitive)")

            # data source subframe
            atc_data_source_frm = Frame(atc_data_frm, style="primary.TFrame")
            atc_data_source_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")
            atc_data_source_label = Label(atc_data_source_frm, text="Data source:",
                                          style="primary.TLabel", db_w=True, width=12)
            atc_data_source_label.pack(side="left", padx=2)
            atc_data_source = Combobox(atc_data_source_frm,
                                       textvariable=self.atc_data_source_var,
                                       values=self.atc_data_sources,
                                       state="readonly",
                                       width=max([len(_) for _ in self.atc_data_sources]),
                                       db_w=True)
            atc_data_source.pack(side="left", padx=2, fill="x", expand=True)
            create_tooltip(atc_data_source, "Select data source")

        # frame for options
        d_options_frm = Frame(d_frm, style="primary.TFrame")
        d_options_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, anchor="w")

        # general options subframe
        atc_display_options_frm = LabelFrame(d_options_frm, text="General",
                                             style="primary_sub.TLabelframe")
        atc_display_options_frm.pack(fill="both")

        # colors subframe
        atc_display_options_top_frm = Frame(atc_display_options_frm, style="primary.TFrame")
        atc_display_options_top_frm.pack(fill="both", pady=2)

        # set color scale button
        self.color_scale_btn_atc = Button(atc_display_options_top_frm, text="Set Color Scale",
                                          style="primary.TButton", db_w=True, atc_w=True,
                                          command=lambda: ColorScalePopup(self))
        self.color_scale_btn_atc.pack(side="left", padx=2)
        create_tooltip(self.color_scale_btn_atc,
                       self.color_scale_tt_template + self.color_scale_var.get())

        # set border button
        self.show_border_btn_atc = Button(atc_display_options_top_frm, text="Set Border",
                                          style="primary.TButton", db_w=True, atc_w=True,
                                          command=lambda: BorderPopup(self))
        self.show_border_btn_atc.pack(side="left", padx=2)
        create_tooltip(self.show_border_btn_atc, self.show_border_tt_template
                       + "\nCurrent properties: Color: " + self.border_color.get()
                       + ", Width: " + self.border_width.get())

        # display legend checkmark
        self.atc_legend_enable = Checkbutton(atc_display_options_top_frm, text="Legend",
                                             style="primary.TCheckbutton", db_w=True,
                                             atc_w=True, onvalue=True, offvalue=False,
                                             variable=self.atc_legend_enabled_control)
        self.atc_legend_enable.pack(side="right", padx=2)
        create_tooltip(self.atc_legend_enable,
                       "Displays a legend in form of a weighted color bar. "
                       "Disabled for summary plots with specific color propagation enabled.")

        mesh_display_options_bottom_frm = Frame(atc_display_options_frm, style="primary.TFrame")
        mesh_display_options_bottom_frm.pack(fill="both", pady=(0, 2))

        # labels
        atc_label = Combobox(mesh_display_options_bottom_frm,
                             textvariable=self.atc_label_var,
                             state="readonly",
                             width=11,
                             values=["all", "propagation", "drugs", "none"],
                             db_w=True,
                             atc_w=True)
        atc_label.pack(side="right", padx=2)
        create_tooltip(atc_label, "Enables/Disables display of labels inside sunburst wedges")
        atc_label_label = Label(mesh_display_options_bottom_frm, text="Display Labels:",
                                style="primary.TLabel", db_w=True, atc_w=True)
        atc_label_label.pack(side="right", padx=2)

        # wedge width
        atc_wedge_width_label = Label(mesh_display_options_bottom_frm, text="Wedge Width:",
                                      style="primary.TLabel", db_w=True, atc_w=True)
        atc_wedge_width_label.pack(side="left", padx=2)
        atc_wedge_width = Combobox(mesh_display_options_bottom_frm,
                                   textvariable=self.atc_wedge_width_var,
                                   state="readonly",
                                   width=9,
                                   values=["total", "remainder"],
                                   db_w=True,
                                   atc_w=True)
        atc_wedge_width.pack(side="left", padx=2)
        create_tooltip(atc_wedge_width,
                       "Change the sunburst display option from a full outer circle (total)"
                       " to count-based wedge-widths (remainder)")

        # propagation subframe
        atc_propagate_frm = LabelFrame(d_options_frm, text="Propagation",
                                       style="primary_sub.TLabelframe")
        atc_propagate_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="both")
        self.atc_propagate_enabled_control.set(False)
        self.atc_propagate_enable = Checkbutton(atc_propagate_frm,
                                                text="Enable",
                                                style="primary.TCheckbutton",
                                                db_w=True,
                                                atc_w=True,
                                                onvalue=True,
                                                offvalue=False,
                                                variable=self.atc_propagate_enabled_control,
                                                command=partial(self.checkbox_controller,
                                                                "atc_propagate"))
        self.atc_propagate_enable.pack(side="left", padx=2)
        create_tooltip(self.atc_propagate_enable,
                       "Select to enable propagation of counts and colors based on ontology level")
        # propagate color specificity
        self.atc_propagate_color_lbl = Label(atc_propagate_frm, text="Color:",
                                             style="primary.TLabel")
        self.atc_propagate_color_lbl.pack(side="left", padx=2)
        self.atc_propagate_color_lbl.configure(state="disabled")
        create_tooltip(self.atc_propagate_color_lbl, self.propagate_color_tt)
        self.atc_propagate_color = Combobox(atc_propagate_frm,
                                            textvariable=self.atc_propagate_color_var,
                                            state="readonly",
                                            values=["off", "specific", "global"],
                                            width=10)
        self.atc_propagate_color.pack(side="left", padx=2)
        self.atc_propagate_color.configure(state="disabled")
        create_tooltip(self.atc_propagate_color, self.propagate_color_tt)

        # propagate counts
        self.atc_propagate_counts_lbl = Label(atc_propagate_frm, text="Counts:",
                                              style="primary.TLabel")
        self.atc_propagate_counts_lbl.pack(side="left", padx=2)
        self.atc_propagate_counts_lbl.configure(state="disabled")
        create_tooltip(self.atc_propagate_counts_lbl, self.propagate_counts_tt)
        self.atc_propagate_counts = Combobox(atc_propagate_frm,
                                             textvariable=self.atc_propagate_counts_var,
                                             state="readonly",
                                             values=["off", "level", "all"],
                                             width=4)
        self.atc_propagate_counts.pack(side="left", padx=2)
        self.atc_propagate_counts.configure(state="disabled")
        create_tooltip(self.atc_propagate_counts, self.propagate_counts_tt)

        # propagate level
        self.atc_propagate_lvl_lbl = Label(atc_propagate_frm, text="Level: ",
                                           style="primary.TLabel")
        self.atc_propagate_lvl_lbl.pack(side="left", padx=2)
        self.atc_propagate_lvl_lbl.configure(state="disabled")
        self.atc_propagate_lvl = Combobox(atc_propagate_frm,
                                          textvariable=self.atc_propagate_lvl_var,
                                          width=3,
                                          state="readonly",
                                          values=[str(_) for _ in range(0, 6, 1) if _ != 0])
        self.atc_propagate_lvl.pack(side="left", padx=2)
        self.atc_propagate_lvl.configure(state="disabled")
        create_tooltip(self.atc_propagate_lvl, "Propagate colors to defined level "
                                               "from outer to inner levels\n"
                                               "---\n"
                                               "affects color propagation when set to "
                                               "'specific' or 'global'\n"
                                               "affects count propagation when set to 'level'"
                                               "ALT:Enable 'Propagation' to modify "
                                               "propagation level")

        # summary plot subframe
        atc_summary_plot_frm = LabelFrame(d_options_frm, text="Plot",
                                          style="primary_sub.TLabelframe")
        atc_summary_plot_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="both")

        # checkbutton to toggle overview / detailed view
        self.atc_summary_plot_cols = Entry(atc_summary_plot_frm, width=2, validate="focusout",
                                           validatecommand=partial(self.overview_entry_validation,
                                                                   "atc"))
        self.atc_summary_plot_cols.insert(0, "5")
        self.atc_summary_plot = Checkbutton(atc_summary_plot_frm,
                                            text="Summary",
                                            style="primary.TCheckbutton",
                                            db_w=True,
                                            atc_w=True,
                                            variable=self.atc_summary_plot_control,
                                            onvalue=True,
                                            offvalue=False,
                                            command=partial(self.checkbox_controller,
                                                            "atc_summary_plot"))
        self.atc_summary_plot.pack(side="left", padx=2)
        self.atc_summary_plot_lbl = Label(atc_summary_plot_frm, text="Columns: ",
                                          style="primary.TLabel")
        self.atc_summary_plot_lbl.pack(side="left", padx=2)
        self.atc_summary_plot_cols.pack(side="left", padx=2)
        create_tooltip(self.atc_summary_plot_cols,
                       "Enter amount of columns in range (1..20)ALT:Enable 'Summary Plot' "
                       "to modify amount of columns")
        create_tooltip(self.atc_summary_plot,
                       "Select to plot all data in a combined overview (resource intensive, "
                       "set Labels to 'none' for faster loading)")

        # run buttons frame
        d_run_frm = Frame(d_frm, style="primary.TFrame")
        d_run_frm.pack(ipadx=2, ipady=2, padx=2, pady=2, fill="x")

        # plot
        atc_plot = Button(d_run_frm, text="Plot", style="primary.TButton",
                          command=partial(self.plot, "atc"),
                          db_w=True, atc_w=True)
        atc_plot.pack(side="right", padx=2)
        create_tooltip(atc_plot, self.plot_tt_template)

        # save plot button
        save_plot_btn = Checkbutton(d_run_frm, text="Save", style="primary.TCheckbutton",
                                    variable=self.export_plot_var, db_w=True, atc_w=True)
        save_plot_btn.pack(side="right", padx=2)
        create_tooltip(save_plot_btn, self.save_plot_tt_template)

        if db_functions:
            # export
            atc_export = Button(d_run_frm, text="Export", style="primary.TButton",
                                command=self.atc_export, db_w=True)
            atc_export.pack(side="right", padx=2)
            create_tooltip(atc_export, self.export_tt_template)

    def checkbox_controller(self, mode: str = None):
        """Toggles widgets based on state of a BooleanVar

        :param mode: must be in ['mesh_summary_plot', 'mesh_propagate', 'atc_summary_plot',
                                 'atc_propagate']
        """
        checkbox, toggle_widgets, var_entry_int_pairs = None, [], []
        if mode == "atc_summary_plot":
            checkbox = self.atc_summary_plot_control
            toggle_widgets = [self.atc_summary_plot_cols, self.atc_summary_plot_lbl]
            var_entry_int_pairs = [(self.atc_summary_plot_var, self.atc_summary_plot_cols)]
        elif mode == "atc_propagate":
            checkbox = self.atc_propagate_enabled_control
            toggle_widgets = [self.atc_propagate_lvl_lbl, self.atc_propagate_lvl,
                              self.atc_propagate_color_lbl, self.atc_propagate_color,
                              self.atc_propagate_counts_lbl, self.atc_propagate_counts]
        elif mode == "mesh_summary_plot":
            checkbox = self.mesh_summary_plot_control
            toggle_widgets = [self.mesh_summary_plot_cols, self.mesh_summary_plot_lbl]
            var_entry_int_pairs = [(self.mesh_summary_plot_var, self.mesh_summary_plot_cols)]
        elif mode == "mesh_propagate":
            checkbox = self.mesh_propagate_enabled_control
            toggle_widgets = [self.mesh_propagate_lvl_lbl, self.mesh_propagate_lvl,
                              self.mesh_propagate_color_lbl, self.mesh_propagate_color,
                              self.mesh_propagate_counts_lbl, self.mesh_propagate_counts]

        for widget in toggle_widgets:
            if checkbox.get():
                widget.configure(state="readonly" if isinstance(widget, Combobox) else "normal")
            else:
                widget.configure(state="disabled")

        for var_entry in var_entry_int_pairs:
            if checkbox.get():
                var_entry[0].set(int(var_entry[1].get()))
            else:
                var_entry[0].set(0)

        if not checkbox.get():
            self.set_status("")

    def toggle_checkbox_widgets(self, mode: str = None, enable: bool = None):
        """Toggles all checkbox related widgets (not part of toggle_widgets routine)

        :param mode: Must be in ['mesh', 'atc']
        :param enable: True to enable, False to disable
        """
        controller_widgets = None
        if mode == "mesh":
            controller_widgets = [self.mesh_propagate_enable, self.mesh_summary_plot]
        elif mode == "atc":
            controller_widgets = [self.atc_propagate_enable, self.atc_summary_plot]
        elif mode == "db":
            controller_widgets = [self.atc_propagate_enable, self.atc_summary_plot,
                                  self.mesh_propagate_enable, self.mesh_summary_plot]

        if enable:
            for widget in controller_widgets:
                widget.configure(state="normal")

            # call checkbox controller for respective toggling of children
            if mode == "mesh":
                self.checkbox_controller(mode="mesh_summary_plot")
                self.checkbox_controller(mode="mesh_propagate")
            elif mode == "atc":
                self.checkbox_controller(mode="atc_summary_plot")
                self.checkbox_controller(mode="atc_propagate")

        else:
            # iterate over controller widgets, disable children within the controller widgets frame
            for controller_widget in controller_widgets:
                for child in controller_widget.master.winfo_children():
                    child.configure(state="disabled")

    def overview_entry_validation(self, mode: str = None):
        """Validates Entry used for defining columns for Overview plot

        :param mode: String, must be 'mesh' or 'atc'
        """
        failed = False
        entry, target_var, target_value = None, None, None

        if mode == "atc":
            entry = self.atc_summary_plot_cols
            target_var = self.atc_summary_plot_var
        elif mode == "mesh":
            entry = self.mesh_summary_plot_cols
            target_var = self.mesh_summary_plot_var

        try:
            target_value = int(entry.get())
        except ValueError:
            failed = True

        if not failed and (target_value <= 0 or target_value > 20):
            failed = True

        if failed:
            self.set_status("ERROR - only integers > 0 and < 20 allowed")
            entry.delete(0, END)
            entry.insert(0, "5")
        else:
            target_var.set(target_value)
            self.set_status("")
        return False

    def check_init(self, obj: [MeSHSunburst, ATCSunburst] = None) -> bool:
        """Checks if core objects are initialized

        :param obj: core object
        :returns: True if initialization was successful
        """
        if not os.path.isfile(self.database_var.get()):
            self.set_database()
        if not os.path.isfile(self.database_var.get()):
            messagebox.showerror("Database", "Could not initialize database, "
                                             "'Export' and 'Plot' functionalities disabled")
            return False

        if not obj.is_init:
            if isinstance(obj, MeSHSunburst):
                self.set_status("Initializing MeSH-tree ..")
                self.p.init(self.database_var.get())
            elif isinstance(obj, ATCSunburst):
                self.set_status("Initializing ATC-tree ..")
                self.d.init(self.database_var.get())

        return obj.is_init

    def set_status(self, text: str = None):
        """Set global status in GUI
        :param text: Text to display as status message
        """
        self.status_var.set("\n".join(textwrap.wrap(text, 65)))
        self.update()

    def configure_p(self):
        """Hand over GUI settings to MeSHSunburst object"""
        self.p.set_color_scale(json.loads(self.color_scale_var.get().replace("'", '"')))
        self.p.set_settings({
            "show_border": self.show_border_var.get(),
            "border_color": self.border_color.get(),
            "border_width": float(self.border_width.get()),
            "export_plot": self.export_plot_var.get(),
            "mesh_drop_empty_last_child": self.mesh_drop_empty_var.get(),
            "mesh_propagate_enable": bool(self.mesh_propagate_enabled_control.get()),
            "mesh_propagate_lvl": int(self.mesh_propagate_lvl_var.get()),
            "mesh_propagate_color": self.mesh_propagate_color_var.get(),
            "mesh_propagate_counts": self.mesh_propagate_counts_var.get(),
            "mesh_labels": self.mesh_label_var.get(),
            "mesh_summary_plot": self.mesh_summary_plot_var.get(),
        })

    def configure_d(self):
        """Hand over GUI settings to ATCSunburst object"""
        self.d.set_color_scale(json.loads(self.color_scale_var.get().replace("'", '"')))
        self.d.set_settings({
            "show_border": self.show_border_var.get(),
            "border_color": self.border_color.get(),
            "border_width": float(self.border_width.get()),
            "export_plot": self.export_plot_var.get(),
            "atc_propagate_enable": bool(self.atc_propagate_enabled_control.get()),
            "atc_propagate_lvl": int(self.atc_propagate_lvl_var.get()),
            "atc_propagate_color": self.atc_propagate_color_var.get(),
            "atc_propagate_counts": self.atc_propagate_counts_var.get(),
            "atc_labels": self.atc_label_var.get(),
            "atc_wedge_width": self.atc_wedge_width_var.get(),
            "atc_summary_plot": self.atc_summary_plot_var.get()
        })

    def toggle_widgets(self, enable: bool = None, mode: str = None, dedicated_parent: [Frame, LabelFrame] = None):
        """Enables/disables widgets in GUI"""
        wdgs = (Button, Checkbutton, Entry, Label, Radiobutton)
        if mode == "recent":
            mode = self.recent_ui_toggle_mode
            # additionally toggle checkbox widgets
            self.toggle_checkbox_widgets(mode=mode, enable=enable)
        else:
            self.recent_ui_toggle_mode = mode

        if enable:
            state = "normal"
            combo_state = "readonly"
        else:
            state = combo_state = "disabled"

        if not dedicated_parent:
            parent = self
        else:
            parent = dedicated_parent

        self._toggle_widgets_recursive(parent, state, combo_state, mode, wdgs)
        self.update()

    def _toggle_widgets_recursive(self, parent, state, combo_state, mode, wdgs):
        """Recursive function to set states for widgets at any depth."""

        for child in parent.winfo_children():
            if isinstance(child, wdgs) and self.is_eligible_for_toggle(child=child, mode=mode):
                child.configure(state=state)
            elif isinstance(child, Combobox) and self.is_eligible_for_toggle(child=child, mode=mode):
                child.configure(state=combo_state)
            elif isinstance(child, (LabelFrame, Frame)):
                self._toggle_widgets_recursive(child, state, combo_state, mode, wdgs)

    @staticmethod
    def is_eligible_for_toggle(child: [Frame, Button, Checkbutton, Entry, Label, Radiobutton, Combobox] = None,
                               mode: str = None):
        return mode == "db" and child.db_w or mode == "mesh" and child.mesh_w or mode == "atc" and child.atc_w

    def rollback_ui(self):
        """Removes ATC/MeSH related widgets"""
        self.mesh_frame.destroy()
        self.atc_frame.destroy()
        self.status_frame.destroy()
        self.update()

        self.mesh_frame = Frame(self)
        self.mesh_frame.pack(fill="both")
        self.atc_frame = Frame(self)
        self.atc_frame.pack(fill="both")
        self.status_frame = LabelFrame(self, text="Status", style="dark.TLabelframe")
        self.status_frame.pack(ipadx=2, ipady=2, fill="both")
        status = Label(self.status_frame, textvariable=self.status_var, style="dark.TLabel")
        status.pack(padx=2)
        self.update()

    def set_database(self, db_path: str = None):
        """Prompt to set database, extracts .tar.gz or verifies chosen .db file, sets class variable

        :param db_path: If path to database is given, only integrity is verified without opening a dialog
        """
        if not db_path:
            db_path = filedialog.askopenfilename(title="Database",
                                                 filetypes=[("DrugVision SQLite3 database",
                                                             ".db .tar.gz"),
                                                            ("All files", "*")])

        if not db_path:
            return

        if db_path.endswith(".db"):
            if self.p.verify_db(db_path):
                self.database_var.set(db_path)
                self.build_mesh_ui(db_functions=True)
                self.build_atc_ui(db_functions=True)
                self.update()
                self.toggle_widgets(enable=True, mode="db")
                self.atc_file_loaded = ""
                self.mesh_file_loaded = ""
            else:
                messagebox.showerror("Database", f"Database {db_path} could not be verified.")

        elif db_path.endswith(".tar.gz"):
            messagebox.showinfo("Database", "Unpacking database archive .. "
                                            "GUI will be unresponsive until finished")
            with tarfile.open(db_path, "r:gz") as tar:
                print(f"Extracting {db_path} ..")
                names = tar.getnames()
                tar.extractall()
            db_path = db_path.rstrip(".tar.gz") + ".db"
            if os.path.isfile(db_path):
                messagebox.showinfo("Database", f"Successfully extracted {db_path}")
                self.set_database(db_path)
            else:
                messagebox.showwarning("Database", f"Extracted unknown file: {names[0]} - "
                                                   f"please select the database manually")
                self.set_database()

    @exception_as_popup
    def plot(self, mode: str = None):
        """Populate tree, update settings of core objects, plot and prompt to overwrite file
        if Excel was loaded

        :param mode: Must be in 'atc', 'mesh'
        """
        populate_tsv = None
        populate_excel = None
        populate_data_source = None
        input_fn = None
        obj = None
        configure = None
        asset = None
        datasource = None
        legend = None
        cfg_exclude = ""
        plot_type = None

        # assign variables based on mode
        if mode == "atc":
            input_fn = self.atc_file_loaded
            populate_data_source = self.d.populate_atc_from_data_source
            populate_tsv = self.d.populate_atc_from_tsv
            populate_excel = self.d.load_atc_excel
            obj = self.d
            configure = self.configure_d
            asset = self.atc_asset_var.get()
            datasource = self.atc_data_source_var.get()
            legend = self.atc_legend_enabled_control.get()
            cfg_exclude = "mesh_"
            plot_type = None
        elif mode == "mesh":
            input_fn = self.mesh_file_loaded
            populate_data_source = self.p.populate_mesh_from_data_source
            populate_tsv = self.p.populate_mesh_from_tsv
            populate_excel = self.p.load_mesh_excel
            obj = self.p
            configure = self.configure_p
            asset = self.mesh_asset_var.get()
            datasource = self.mesh_data_source_var.get()
            legend = self.mesh_legend_enabled_control.get()
            cfg_exclude = "atc_"
            plot_type = self.mesh_plot_type_var.get()

        obj.s["legend"] = legend
        obj.s["plot_type"] = plot_type

        # populate tree from Excel or database data
        if asset == "CUSTOM":
            self.set_status("Populating custom tree ..")
        else:
            self.set_status(f"Populating {mode.upper()} tree ..")
        if input_fn:
            if os.path.splitext(input_fn)[-1] == ".tsv":
                if datasource == "TSV file":
                    populate_tsv(input_fn)
                elif datasource.startswith("custom_sep_"):
                    self.p.populate_custom_ontology_from_tsv(fn=input_fn, ontology_type=datasource)
                else:
                    self.p.custom_ontology = build_non_separator_based_tree(file_name=input_fn,
                                                                            float_sep=self.custom_ontology_separator)
                    self.p.custom_ontology_title = os.path.abspath(input_fn).split(os.sep)[-1]
                    self.p.populate_custom_ontology_from_web()
            else:
                populate_excel(input_fn, read_settings=False, populate=True)
        else:
            # in case custom ontology was loaded, no file is specified
            if self.p.custom_ontology:
                self.p.populate_custom_ontology_from_web()
            else:
                # otherwise, verify db is loaded and populate tree
                if not self.check_init(obj):
                    return
                populate_data_source(asset, datasource)

        # update settings of core object based on current GUI configuration
        configure()

        # show warning once in case labels and summary plot is enabled
        if not self.performance_warning_shown and obj.s[f"{mode}_summary_plot"] \
                and obj.s[f"{mode}_labels"] == "all":
            messagebox.showwarning(title="Performance of plot",
                                   message="Displaying very large ontologies as a summary plot and "
                                           "displaying all associated labels may result in longer "
                                           "loading times and less responsive interaction in the "
                                           "browser.")
            self.performance_warning_shown = True

        # launch plot creation as thread
        thread = Thread(target=obj.plot, args=())
        thread.start()

        # disable UI
        self.toggle_widgets(enable=False, mode="recent")

        # refresh status based on thread responses
        while thread.is_alive():
            self.set_status(obj.thread_status)
            time.sleep(0.1)

        # enable UI, set final status
        self.toggle_widgets(enable=True, mode="recent")
        if obj.plot_error:
            self.set_status("Failed to display incompatible ontology")
            raise Exception(obj.plot_error)

        self.set_status("Plot displayed in browser")

        # in case self.export_plot_var.get() is checked (sets config param "export_plot"),
        # thread will return generated .html and .tsv file paths
        thread_ret = obj.thread_return
        if thread_ret:
            html_path, tsv_path = thread_ret

            # show both filenames as info popup
            messagebox.showinfo("Export Plot", f"Exported plot to:\n{html_path}\n\nExported "
                                               f"template to:\n{tsv_path}")

        # prompt to overwrite Excel file if settings changed since load
        if input_fn and input_fn.endswith(".xlsx"):
            accepted = ["color_scale", "show_border", "border_color", "border_width"]
            accepted.extend([_ for _ in self.loaded_settings.keys() if _.startswith(mode)])
            modified = {k: (str(self.loaded_settings[k]), str(obj.s[k]))
                        for k in self.loaded_settings.keys()
                        if str(obj.s[k]) != str(self.loaded_settings[k]) and k in accepted}
            tmp = "\n".join([f"{k}: '{v[0]}' -> '{v[1]}'" for k, v in modified.items()])
            if modified:
                overwrite = messagebox.askokcancel(title="Settings changed",
                                                   message=f"Settings changed, "
                                                           f"overwrite {input_fn} ?\n\n{tmp}")
                if overwrite:
                    settings = [(k, v) for k, v in obj.s.items()
                                if not k.startswith(cfg_exclude) and k != "default_color"]
                    out_fn = obj.export_settings(fn=input_fn, settings=settings)
                    self.set_status(f"Updated {out_fn}")

    @exception_as_popup
    def mesh_export(self):
        """Update settings of core object, populate and export MeSH-tree to Excel"""
        if not self.check_init(self.p):
            return

        self.configure_p()

        self.set_status("Populating MeSH-tree ..")
        self.p.populate_mesh_from_data_source(drug_name=self.mesh_asset_var.get(),
                                              data_source=self.mesh_data_source_var.get())

        export_popup = ExportPopup(self, "Export as",
                                   "Export MeSH-Tree as Excel (remembers settings) or .tsv file")
        selection = export_popup.selection
        export_as_template = export_popup.export_as_template.get()

        if selection:
            self.set_status(f"Exporting MeSH-tree to {selection} ..")
            export_fn = self.p.export_mesh_tree(mode=selection, template=export_as_template)
            messagebox.showinfo("Export", f"Exported MeSH-tree to: {export_fn}")

        self.set_status("")

    @exception_as_popup
    def atc_export(self):
        """Update settings of core object, populate and export ATC-tree to Excel"""
        if not self.check_init(self.d):
            return

        self.configure_d()

        self.set_status("Populating ATC-tree ..")
        self.d.populate_atc_from_data_source(phenotype_name=self.atc_asset_var.get(),
                                             data_source=self.atc_data_source_var.get())

        export_popup = ExportPopup(self, "Export as",
                                   "Export ATC-Tree as Excel (remember settings) or .tsv file")
        selection = export_popup.selection
        export_as_template = export_popup.export_as_template.get()

        if selection:
            self.set_status(f"Exporting ATC-tree to {selection} ..")
            export_fn = self.d.export_atc_tree(mode=selection, template=export_as_template)
            messagebox.showinfo("Export", f"Exported ATC-tree to: {export_fn}")

        self.set_status("")

    @exception_as_popup
    def load_url(self):
        """Download and visualize a .obo ontology"""

        self.obo = SelectOptionsPopup(
            parent=self,
            title="Choose Ontology",
            info_text="Select Ontology to download and visualize,\nor define a custom .obo URL",
            is_ontology_popup=True,
            options={
                "hpo": ("Human Phenotype Ontology", self.hpo_ontology_tt),
                "go_mf": ("Gene Ontology (molecular function)",
                          "Fetches the GeneOntology (namespace: molecular_function) "
                          f"{self.gene_ontology_tt}"),
                "go_cp": ("Gene Ontology (cellular component)",
                          "Fetches the GeneOntology (namespace: cellular_component) "
                          f"{self.gene_ontology_tt}"),
                "go_bp": ("Gene Ontology (biological process)",
                          "Fetches the GeneOntology (namespace: biological_process) "
                          f"{self.gene_ontology_tt}"),
                # "po": ("Plant Ontology", self.po_ontology_tt),
                "cl": ("Cell Ontology", self.cl_ontology_tt),
                # "chebi": ("CHEBI Ontology", self.chebi_ontology_tt),
                # "uberon": ("Uberon Anatomy Ontology", self.uberon_ontology_tt),
                "doid": ("Human Disease Ontology", self.doid_ontology_tt),
                "custom_url": ("Custom URL", "Specify a custom .obo url"),
            },
        )

        description = self.obo.description
        ontology = self.obo.result
        if not ontology:
            self.set_status("Aborted ontology download")
            return

        self.rollback_ontology_variables()
        self.p.custom_ontology = get_remote_ontology(ontology_short=ontology, app=self)
        self.p.custom_ontology_title = description

        # set core object settings, assign functions, set status, rollback ui
        self.rollback_ui()
        self.change_theme_color(foreground=self.d4_black, background=self.d4_custom)
        self.build_mesh_ui(db_functions=False)
        self.mesh_label_var.set("none")  # hide labels
        self.mesh_legend_enabled_control.set(False)  # disable legend
        self.mesh_data_source_var.set(ontology)
        self.mesh_asset_var.set(description)
        self.title(f"OntoloViz - {description}")
        self.reset_load_button_styles()
        self.recent_ui_toggle_mode = "mesh"
        self.update()

    @exception_as_popup
    def load_file(self):
        """Prompt to load Excel/.tsv file

        If file ends on .xlsx: verify file and get tree type based on option in 'Settings'
            and number of columns in 'Tree' tabs, enable/disable and configure respective widgets
        """
        input_fn = filedialog.askopenfilename(filetypes=[("Tree Table", ".xlsx .tsv"),
                                                         ("SQLite3 database", ".db .tar.gz"),
                                                         ("All files", "*")],
                                              title="Load ontology from file")
        if not input_fn:
            return

        self.rollback_ontology_variables()
        obj = None

        if input_fn.endswith(".db") or input_fn.endswith(".tar.gz"):
            self.set_database(input_fn)
            return

        # verify file
        tree_type = self.p.verify_file(input_fn)
        custom_ontology = None
        if not tree_type:
            if input_fn.endswith("xlsx"):
                messagebox.showerror(title="File format not supported",
                                     message="The input file could not be automatically identified,"
                                             " custom ontologies are only supported for .tsv files."
                                             " Convert your data to .tsv, and follow the guidelines"
                                             " from https://github.com/Delta4AI/OntoloViz")
                return
            _custom_ontology = SelectOptionsPopup(
                parent=self, title="Choose Ontology Type",
                info_text="The ontology type could not be detected automatically. "
                          "What type of ontology are you trying to import? "
                          "Find out about the supported structures at "
                          "https://github.com/Delta4AI/OntoloViz",
                options={
                    "custom_sep_dot": ("Dot-separated", "e.g. MeSH: 'C01.001.002'"),
                    "custom_sep_slash": ("Slash-separated", "e.g. C01/001/002"),
                    "custom_sep_colon": ("Colon-separated", "e.g. C01,001,002"),
                    "custom_sep_underscore": ("Underscore-separated", "e.g. C01_001_002"),
                    "custom_non_sep": ("Parent-based",
                                       "Unstructured ontologies that do not follow a structured "
                                       "schema, e.g. HPO IDs in the format: HP:0001300\nRequires "
                                       "6 column layout and defined parent-id for each node:\n"
                                       "-------------------------------------------------\n"
                                       "ID | Parent | Label | Description | Count | Color")
                }
            )
            custom_ontology = _custom_ontology.description
            tree_type = _custom_ontology.result
            self.custom_ontology_separator = _custom_ontology.separator
            if not tree_type:
                self.set_status("Aborted file loading")
                return

        # set core object settings, assign functions, set status, rollback ui
        self.rollback_ui()
        self.set_status("")
        if tree_type.startswith("atc"):
            self.set_status("Loading ATC tree from file ..")
            self.change_theme_color(foreground=self.d4_white, background=self.d4_red)
            self.build_atc_ui(db_functions=False)
            self.update()
            if tree_type == "atc_excel":
                self.d.load_atc_excel(fn=input_fn, read_settings=True, populate=False)
                self.atc_data_source_var.set("Excel file")
            elif tree_type == "atc_tsv":
                self.atc_data_source_var.set("TSV file")
            self.atc_asset_var.set(self.d.phenotype_name)
            obj = self.d
        elif tree_type.startswith("mesh"):
            self.set_status("Loading MeSH tree from file ..")
            self.change_theme_color(foreground=self.d4_white, background=self.d4_purple)
            self.build_mesh_ui(db_functions=False)
            self.update()
            if tree_type == "mesh_excel":
                self.p.load_mesh_excel(fn=input_fn, read_settings=True, populate=False)
                self.mesh_data_source_var.set("Excel file")
            elif tree_type == "mesh_tsv":
                self.mesh_data_source_var.set("TSV file")
            self.mesh_asset_var.set(self.p.drug_name)
            obj = self.p
        else:
            # custom ontologies
            self.set_status("Loading custom ontology from file ..")
            self.change_theme_color(foreground=self.d4_black, background=self.d4_custom)
            self.build_mesh_ui(db_functions=False)
            self.update()
            self.mesh_data_source_var.set(tree_type)
            self.mesh_asset_var.set("CUSTOM")
            obj = self.p

        # disable all widgets to be enabled later
        self.toggle_widgets(enable=False, mode="db")

        # set general settings in GUI
        self.color_scale_var.set(str(obj.s["color_scale"]))
        create_tooltip(self.color_scale, self.color_scale_tt_template + self.color_scale_var.get())
        self.show_border_var.set(obj.s["show_border"])
        self.border_color.set(obj.s["border_color"])
        self.border_width.set(str(obj.s["border_width"]))
        create_tooltip(self.show_border, self.show_border_tt_template
                       + "\nCurrent properties: Color: " + self.border_color.get()
                       + ", Width: " + self.border_width.get())
        self.export_plot_var.set(obj.s["export_plot"])

        # set specific settings in GUI
        if tree_type.startswith("atc"):
            self.toggle_widgets(enable=True, mode="atc")
            self.atc_file_loaded = input_fn
            self.set_status(f"ATC tree loaded: {input_fn}")
            self.title("OntoloViz - ATC Ontology")
            if tree_type == "atc_excel":
                self.atc_label_var.set(obj.s["atc_labels"])
                self.atc_wedge_width_var.set(obj.s["atc_wedge_width"])
        elif tree_type.startswith("mesh"):
            self.toggle_widgets(enable=True, mode="mesh")
            self.set_status(f"MeSH tree loaded: {input_fn}")
            self.mesh_file_loaded = input_fn
            self.title("OntoloViz - MeSH Ontology")
            if tree_type == "mesh_excel":
                self.mesh_drop_empty_var.set(obj.s["mesh_drop_empty_last_child"])
                self.mesh_label_var.set(obj.s["mesh_labels"])
        else:
            # custom ontologies
            self.toggle_widgets(enable=True, mode="mesh")
            self.set_status(f"Custom tree loaded: {input_fn}")
            self.mesh_file_loaded = input_fn
            self.title(f"OntoloViz - {custom_ontology} Ontology")

        # store settings to check later if they have been modified if Excel was loaded
        if tree_type.endswith("_excel"):
            self.loaded_settings = {k: v for k, v in obj.s.items()}

        # reset button style
        self.reset_load_button_styles()

    def reset_load_button_styles(self):
        """Resets the styles of the load file buttons and removes green outline"""
        self.load_file_btn.configure(style="dark.TButton")
        self.load_obo_url_btn.configure(style="dark.TButton")

    def rollback_ontology_variables(self):
        """Rolls back some of the variables for loading a new file properly"""
        self.atc_file_loaded = ""
        self.mesh_file_loaded = ""
        self.p.custom_ontology = None
        self.p.custom_ontology_title = None
        self.loaded_settings = {}
        self.custom_ontology_separator = None


def run_app():
    """Creates instance of App, launches mainloop"""
    print("Launching OntoloViz GUI ..")
    app = App()
    app.mainloop()


if __name__ == "__main__":
    run_app()



import requests
import re
from typing import Union

zero = 0.000001337
fake_one = 1.000001337
white = "#FFFFFF"


def build_non_separator_based_tree(file_name: str = None, float_sep: str = None) -> dict:
    """Parse an ontology with child- and parent-ids from a file and build tree structure

    :param file_name: tab separated file with 6 columns:
        id, parent, label, description, count, color
    :param float_sep: if given, counts are considered as floating point values and converted based on given separator
    """
    tree, to_process = parse_file_to_extract_root_nodes_and_processable_lines(file_name, float_sep)

    while True:
        drop_idxs = handle_and_assign_nodes(to_process, tree)
        for idx in sorted(list(set(drop_idxs)), reverse=True):
            del to_process[idx]

        if not to_process:
            break

        print(f"Dropped: {len(drop_idxs)}, Left to process: {len(to_process)}")

    if float_sep:
        print("Normalizing float counts to int")
        normalize_tree_counts_from_float_to_int(tree)

    return tree


def handle_and_assign_nodes(to_process: list = None, tree: dict = None) -> list:
    drop_idxs = []
    for idx, (attempts, node) in enumerate(to_process):
        if attempts >= 20:
            print(f"Dropping node because no suitable parent was found after "
                  f"20 attempts: {node['id']}")
            drop_idxs.append(idx)
            continue

        for sub_tree_id, sub_tree in tree.items():
            parent = node["parent"]
            if parent in sub_tree.keys():
                node["level"] = tree[sub_tree_id][parent]["level"] + 1
                tree[sub_tree_id][node["id"]] = node
                drop_idxs.append(idx)

        to_process[idx][0] += 1
    return drop_idxs


def parse_file_to_extract_root_nodes_and_processable_lines(input_file: str = None, float_sep: str = None) -> tuple:
    tree = dict()
    to_process = list()
    duplicate_check = list()
    with open(file=input_file, mode="r", encoding="utf-8") as f_in:
        for line_idx, line in enumerate(f_in):
            if line_idx == 0:
                continue
            node_ids_unformatted, *line_data = line.rstrip("\n").split("\t")
            node_ids = node_ids_unformatted.split("|")
            for node_id in node_ids:
                original_node_id = node_id
                duplicate_count = duplicate_check.count(node_id) + 1
                if duplicate_count > 1:
                    node_id = f"{node_id}_{duplicate_count}"
                duplicate_check.append(original_node_id)
                handle_and_assign_root_nodes(node_id, tree, to_process, line_data, float_sep)

    return tree, to_process


def handle_and_assign_root_nodes(node_id: str = None, tree: dict = None, to_process: list = None,
                                 line_data: list = None, float_sep: str = None):
    parent = line_data[0]
    count = safe_convert_count(line_data[3], float_sep)
    color = line_data[4]

    node = {
        "id": node_id,
        "parent": parent,
        "level": 0,
        "label": line_data[1],
        "description": line_data[2],
        "counts": count if count != 0 else zero,
        "imported_counts": count if count != 0 else fake_one,
        "color": color if color else white
    }

    # populate first level of tree structure
    if not parent:
        tree[node_id] = {
            node_id: node
        }
    else:
        to_process.append([0, node])


def safe_convert_count(count_as_str: str = None, float_sep: str = None) -> Union[int, float]:
    def_value = 0.0 if float_sep else 0
    try:
        if float_sep:
            return float(count_as_str.replace(float_sep, "."))
        else:
            return int(count_as_str)
    except ValueError:
        return def_value


def normalize_tree_counts_from_float_to_int(tree: dict = None):
    max_val = find_max(tree)
    for main_node in tree:
        for sub_node in tree[main_node]:
            count = tree[main_node][sub_node]["counts"]
            if count != zero and count != fake_one:
                tree[main_node][sub_node]["counts"] = normalize_to_int(count, max_val)


def find_max(tree: dict):
    max_val = zero
    for main_node in tree:
        for sub_node in tree[main_node]:
            if tree[main_node][sub_node]["counts"] > max_val:
                max_val = tree[main_node][sub_node]["counts"]
    return max_val


def normalize_to_int(original_value: float = None, max_float: float = None, max_integer_value: int = 100) -> int:
    return int((original_value / max_float) * max_integer_value)


def get_remote_ontology(ontology_short: str = None, app: object = None, url: str = None,
                        root_id: str = None, min_node_size: int = None) -> dict:
    """Wrapper to get .obo ontologies based on identifiers"""
    if ontology_short == "hpo":
        return build_tree_from_obo_ontology(url="https://purl.obolibrary.org/obo/hp.obo",
                                            descriptor="Human phenotype ontology",
                                            root_id="HP:0000118",
                                            app=app)
    elif ontology_short in ["go_mf", "go_cp", "go_bp"]:
        raw_terms = parse_obo_file(url="https://current.geneontology.org/ontology/go.obo",
                                   descriptor="GeneOntology",
                                   app=app)
        # get GO root IDs and labels:
        root_terms = {v["label"]: v["id"] for k, v in raw_terms.items() if not v["is_a"]}
        root_term_translation = {
            "go_mf": "molecular_function",
            "go_cp": "cellular_component",
            "go_bp": "biological_process"
        }
        root_id = root_terms[root_term_translation[ontology_short]]
        return build_tree_from_obo_ontology(raw_terms=raw_terms,
                                            root_id=root_id,
                                            descriptor="GeneOntology",
                                            app=app,
                                            min_node_size=2)
    elif ontology_short == "po":
        return build_tree_from_obo_ontology(url="https://purl.obolibrary.org/obo/po.obo",
                                            descriptor="Plant Ontology",
                                            app=app,
                                            root_id="PO:0009011",
                                            min_node_size=5)
    elif ontology_short == "cl":
        return build_tree_from_obo_ontology(url="https://purl.obolibrary.org/obo/cl/cl-basic.obo",
                                            descriptor="Cell Ontology",
                                            app=app, min_node_size=2)
    elif ontology_short == "chebi":
        return build_tree_from_obo_ontology(
            url="https://purl.obolibrary.org/obo/chebi/chebi_lite.obo",
            descriptor="CHEBI Ontology",
            app=app,
            root_id="CHEBI:23367")
    elif ontology_short == "uberon":
        return build_tree_from_obo_ontology(url="https://purl.obolibrary.org/obo/uberon/basic.obo",
                                            descriptor="Uberon Anatomy Ontology",
                                            app=app,
                                            root_id="UBERON:0000061",
                                            min_node_size=2)
    elif ontology_short == "doid":
        return build_tree_from_obo_ontology(url="https://purl.obolibrary.org/obo/doid.obo",
                                            descriptor="Human Disease Ontology",
                                            app=app,
                                            root_id="DOID:4")
    elif ontology_short == "custom_url":
        return build_tree_from_obo_ontology(url=app.obo.custom_url,
                                            descriptor=app.obo.description,
                                            app=app,
                                            root_id=app.obo.root_id,
                                            min_node_size=app.obo.min_node_size)


def parse_obo_file(url: str = None, descriptor: str = None, app: object = None,
                   exclude_obsolete_terms: bool = True) -> dict:
    """ Downloads and parses an .obo file

    :param url: url of .obo file
    :param descriptor: descriptor used to show status in app
    :param app: tkinter App object
    :param exclude_obsolete_terms: if True, terms with "is_obsolete: true" will be excluded
    :return: dictionary containing raw parsed obo data
    """
    if app:
        app.set_status(f"Downloading {descriptor} ..")
    response = requests.get(url=url, stream=False)
    response.raise_for_status()

    # Read the response content in chunks
    chunks = []
    for chunk in response.iter_content(chunk_size=128 * 512):
        if app:
            app.set_status(f"Downloading {descriptor} .. {round(len(chunks) * 128 / 2048, 2)} MB")
        chunks.append(chunk.decode("utf-8"))

    # Concatenate the chunks into a single string
    text = ''.join(chunks).split("\n")

    # remove header block
    text = text[text.index("") + 1:]

    # parse and prepare data
    raw_terms = {}
    new_entity = None
    for line in text:
        if line == "[Term]":
            new_entity = {
                "id": None,
                "label": None,
                "namespace": None,
                "def": None,
                "comment": None,
                "is_obsolete": False,
                "xrefs": [],
                "is_a": [],
                "disjoint_from": [],
                "synonyms": [],
            }
        elif line.startswith("id: "):
            new_entity["id"] = line.replace("id: ", "")
        elif line.startswith("name: "):
            new_entity["label"] = line.replace("name: ", "")
        elif line.startswith("def: "):
            new_entity["def"] = line.replace('def: "', "").split('" [')[0]
        elif line.startswith("comment: "):
            new_entity["comment"] = line.replace("comment: ", "")
        elif line.startswith("xref: "):
            new_entity["xrefs"].append(line.replace("xref: ", ""))
        elif line.startswith("is_a: "):
            new_entity["is_a"].append(line.replace("is_a: ", "").split(" ! "))
        elif line.startswith("disjoint_from: "):
            new_entity["disjoint_from"].append(line.replace("disjoint_from: ", "").split(" ! "))
        elif line.startswith("namespace: "):
            new_entity["namespace"] = line.replace("namespace: ", "")
        elif line.startswith("synonym: "):
            new_entity["synonyms"].append(line.replace("synonym: ", "").lstrip('"').split('" '))
        elif line == "is_obsolete: true":
            new_entity["is_obsolete"] = True
        elif line == "":
            if exclude_obsolete_terms and new_entity["is_obsolete"]:
                continue
            raw_terms[new_entity["id"]] = new_entity

    return raw_terms


def build_tree_from_obo_ontology(url: str = None,
                                 descriptor: str = None,
                                 root_id: str = None,
                                 raw_terms: dict = None,
                                 app: object = None,
                                 min_node_size: int = None) -> dict:
    """Downloads and parses the HPO from a remote obo file

    :param url: url of hp.obo file
    :param descriptor: Descriptor for status prints
    :param root_id: ID of "Phenotypic abnormality" to mark starting-point for creation of subtrees
    :param raw_terms: pre-processed terms (skips download/parsing)
    :param app: App object used for status updates
    :param min_node_size: only keeps subtrees if node amount is greater than given value
    :return: dictionary containing tree with necessary parameters to plot
    """

    if not raw_terms:
        raw_terms = parse_obo_file(url=url, descriptor=descriptor, app=app)

    # build first level
    tree = {}
    if root_id:
        for node_id, val in raw_terms.items():
            for is_a in val["is_a"]:
                if root_id == is_a[0]:
                    tree[node_id] = {}
                    tree[node_id][node_id] = val
                    tree[node_id][node_id]["level"] = 0
                    tree[node_id][node_id]["parent"] = ""
    else:
        root_terms = [_["id"] for _ in raw_terms.values() if not _["is_a"]]
        for root_term in root_terms:
            tree[root_term] = {}
            tree[root_term][root_term] = raw_terms[root_term]
            tree[root_term][root_term]["level"] = 0
            tree[root_term][root_term]["parent"] = ""

    # propagate
    iterations = 0
    while True:
        had_content = False
        for sub_tree_id, sub_tree in tree.items():
            for node_id, node in raw_terms.items():
                for is_a in node["is_a"]:
                    is_a_id = is_a[0]
                    if is_a_id in sub_tree.keys() and node_id not in sub_tree.keys():
                        sub_tree[node_id] = {key: value for key, value in node.items()}
                        sub_tree[node_id]["level"] = sub_tree[is_a_id]["level"] + 1
                        sub_tree[node_id]["parent"] = is_a_id
                        had_content = True
        if not had_content:
            break
        else:
            iterations += 1
            if app:
                app.set_status(f"Building {descriptor} tree .. iteration #{iterations}")

    # just in case - clean nodes where parent doesn't exist
    cleaning_iterations = 0
    while True:
        nodes_with_missing_parent = []
        for sub_tree_id, sub_tree in tree.items():
            for key, val in sub_tree.items():
                if val["parent"] and val["parent"] not in sub_tree.keys():
                    nodes_with_missing_parent.append((sub_tree_id, key))

        if not nodes_with_missing_parent:
            break

        cleaning_iterations += 1
        print(f"Cleaning iteration {cleaning_iterations}")
        for sub_tree_id, node_id in nodes_with_missing_parent:
            del tree[sub_tree_id][node_id]

    # add zero counts, color and description
    for sub_tree in tree.values():
        for node in sub_tree.values():
            node["imported_counts"] = 1.000001337
            node["counts"] = 0.000001337
            node["color"] = "#FFFFFF"
            node["description"] = f"Definition: {node['def']}\nComment: {node['comment']}"

    if app:
        app.set_status(f"Parsed {descriptor}")

    if min_node_size:
        sub_tree_ids_to_drop = []
        for sub_tree_id, sub_tree in tree.items():
            if len(sub_tree) < min_node_size:
                sub_tree_ids_to_drop.append(sub_tree_id)
        for sub_tree_id in sub_tree_ids_to_drop:
            del tree[sub_tree_id]

    return tree


def sanitize_string(filename):
    # Define a regular expression pattern for illegal characters and newline
    illegal_chars_pattern = r'[<>:"/\\|?*\x00-\x1F\n]'

    # Replace illegal characters with an underscore
    sanitized_filename = re.sub(illegal_chars_pattern, '_', filename)

    return sanitized_filename



import json
from re import match
from functools import partial
from traceback import format_exc

from tkinter import Toplevel, messagebox, ttk, StringVar, BooleanVar, IntVar, END
from tkinter.ttk import LabelFrame, Frame
from tkinter import Label as LabelOG, Entry as EntryOG
from tkinter.colorchooser import askcolor

from src.ontoloviz.core import MeSHSunburst, ATCSunburst
from src.ontoloviz.core_utils import rgb_to_hex, hex_to_rgb


_key_release = "<KeyRelease>"


class ToggleMixin:
    """Class to add flags to widgets for toggling their state selectively"""
    def __init__(self, db_w=False, mesh_w=False, atc_w=False, **kwargs):
        super().__init__(**kwargs)
        self.db_w = db_w
        self.mesh_w = mesh_w
        self.atc_w = atc_w


class Button(ToggleMixin, ttk.Button):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class Entry(ToggleMixin, ttk.Entry):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class Combobox(ToggleMixin, ttk.Combobox):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class Checkbutton(ToggleMixin, ttk.Checkbutton):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class Label(ToggleMixin, ttk.Label):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class Radiobutton(ToggleMixin, ttk.Radiobutton):
    def __init__(self, master: object = None, db_w: bool = False, mesh_w: bool = False,
                 atc_w: bool = False, **kwargs):
        super().__init__(master=master, db_w=db_w, mesh_w=mesh_w, atc_w=atc_w, **kwargs)


class ToolTip:
    """Class that adds tooltip functionality, call with create_tooltip(widget, text)"""
    def __init__(self, widget: [Label, Checkbutton, Combobox, Entry, Button, Radiobutton] = None):
        self.widget = widget
        self.tip_window = None
        self.id = None
        self.tt_x = self.tt_y = 0

    def showtip(self, text: str = None, alt_text: str = None):
        """Calculate coordinates, create Toplevel, add Label with text to ToolTip"""
        if self.tip_window or not text:
            return

        # calculate coordinates
        tt_x, tt_y, _cx, _cy = self.widget.bbox("insert")
        tt_x = tt_x + self.widget.winfo_rootx() + 57
        tt_y = tt_y + _cy + self.widget.winfo_rooty() + 27

        # create Toplevel
        self.tip_window = tt_window = Toplevel(self.widget)
        tt_window.wm_overrideredirect(True)
        tt_window.wm_geometry(f"+{tt_x}+{tt_y}")

        # add Label with text
        tmp = text if str(self.widget['state']) != "disabled" else alt_text
        label = Label(tt_window, text=tmp, justify="left", relief="solid", borderwidth=0.5,
                      font=("Consolas", 8))
        label.pack(ipadx=1)

    def hidetip(self):
        """Destroy Toplevel of ToolTip"""
        tt_window = self.tip_window
        self.tip_window = None
        if tt_window:
            tt_window.destroy()


def create_tooltip(widget: [Label, Checkbutton, Combobox, Entry, Button, Radiobutton] = None,
                   text: str = None):
    """Create tooltip for any widget.
    Example: create_tooltip(some_widget, text="Test Message")

    :param widget: widget to attach tooltip to
    :param text: Text to display on active widget
    """
    if not widget:
        return
    tool_tip = ToolTip(widget)

    alt_text = "This functionality requires a loaded Excel/TSV file"
    if widget.db_w and not widget.atc_w and not widget.mesh_w:
        alt_text = "This functionality requires a loaded database"
    elif widget.mesh_w and not widget.atc_w:
        alt_text += " containing MeSH data"
    elif widget.atc_w and not widget.mesh_w:
        alt_text += " containing ATC data"

    # individual tooltips when tooltip text contains "ALT:"
    if "ALT:" in text:
        text, alt_text = text.split("ALT:")

    def enter(_event):
        try:
            # add 1 space to beginning of each line and at the end; show tooltip
            tt_text = str(" {}".format(" \n ".join(text.split("\n")))
                          if text.find("\n") != -1 else f" {text} ")
            tt_alt_text = str(" {}".format(" \n ".join(alt_text.split("\n")))
                              if alt_text.find("\n") != -1 else f" {alt_text} ")
            tool_tip.showtip(tt_text, tt_alt_text)
        except Exception as exc:
            print(exc)

    def leave(_event):
        tool_tip.hidetip()

    widget.bind('<Enter>', enter)
    widget.bind('<Leave>', leave)


def update_tooltip(widget: [Label, Checkbutton, Combobox, Entry, Button, Radiobutton] = None,
                   text: str = None):
    """Update the text of TopLevel.Label of an open tooltip
    (if re-creation is not an option, e.g. at checkboxes)
    """
    widget.winfo_children()[0].winfo_children()[0]["text"] = text


def exception_as_popup(func: callable = None):
    """Decorator to show occurring exceptions as popup"""
    def wrapper(*args, **kwargs):
        try:
            ret = func(*args, **kwargs)
            return ret
        except Exception as exc:
            _args = " ".join([str(_) for _ in exc.args])
            messagebox.showerror(f"Error: {_args}", f"{_args}\n\n{'*' * 30}\n\n{format_exc()}")
    return wrapper


class ExportPopup(Toplevel):
    """Popup class with options to export data as Excel, TSV or Cancel"""
    def __init__(self, parent: [MeSHSunburst, ATCSunburst] = None, title: str = None, message: str = None):
        """Export Popup init"""
        super().__init__(parent)
        self.title = title
        self.selection = None
        self.resizable(False, False)
        root = Frame(self)
        root.pack(padx=10, pady=10)

        # create widgets
        Label(root, text=message).pack()
        button_frame = Frame(root)
        button_frame.pack(fill="both", expand=True, pady=(20, 0))
        Button(button_frame, text="Cancel", command=partial(self.select, None)).pack(side="right")
        Button(button_frame, text="TSV", command=partial(self.select, "TSV")).pack(side="right")
        Button(button_frame, text="Excel", command=partial(self.select, "Excel")).pack(side="right")
        self.export_as_template = BooleanVar(value=False)
        check_button = Checkbutton(button_frame, text="Create Template",
                                   variable=self.export_as_template, onvalue=True, offvalue=False)
        check_button.pack(side="right")
        create_tooltip(check_button, "Creates an all-white template with 0 counts")

        # freeze mainloop
        self.wait_window(self)

    def select(self, export_mode: str):
        """Command for buttons 'TSV' and 'Excel'"""
        self.selection = export_mode
        self.destroy()
        return export_mode


class ColorScalePopup(Toplevel):
    """Popup to set the color scale"""
    def __init__(self, parent: [MeSHSunburst, ATCSunburst]):
        """ColorScale Popup init"""
        super().__init__(parent)
        self.title("Set Color Scale")
        self.black = "#000000"
        self.white = "#FFFFFF"
        self.resizable(False, False)
        self.parent = parent
        root = Frame(self)
        root.pack(padx=10, pady=10)

        # informative label
        Label(root,
              text="Enter values for an automatic color scale. The first color defines the default "
                   "color for empty nodes. Requires active propagation to have an "
                   "effect. Thresholds must increase and be in range from 0-100.",
              wraplength=400).pack(pady=(10, 10))

        # header for scale objects
        scale_header = Frame(root)
        scale_header.pack(fill="x", expand=True, pady=5, padx=(33, 175))
        Label(scale_header, text="Threshold [%]").pack(side="left", anchor="w")
        Label(scale_header, text="Hex-Color").pack(side="right")

        # scale objects
        self.scale_frame = Frame(root)
        self.scale_frame.pack(fill="both", expand=True)

        current_scale = json.loads(self.parent.color_scale_var.get().replace("'", '"'))
        self.thresholds = []
        for percentage, hex_color in current_scale:
            self.add_threshold(percentage, hex_color)

        # controller button bar, status
        controller_btn_frm = Frame(root)
        controller_btn_frm.pack(fill="x", expand=True)
        controller_btn_subfrm = Frame(controller_btn_frm)
        controller_btn_subfrm.pack(side="top")
        Button(controller_btn_subfrm, text="-", width=4, command=self.decrease).pack(side="left")
        Button(controller_btn_subfrm, text="+", width=4, command=self.increase).pack(side="left")
        self.status = Label(controller_btn_frm, text="", foreground="red")
        self.status.pack()

        # bottom button bar
        button_frm = Frame(root)
        button_frm.pack(fill="x", expand=True, pady=(20, 0))
        Button(button_frm, text="Cancel", command=lambda: self.destroy()).pack(side="right")
        Button(button_frm, text="Apply", command=self.set).pack(side="right")

        # freeze mainloop
        self.wait_window(self)

    def add_threshold(self, percentage: float, hex_color: str) -> None:
        """Subroutine to create an entry pair inside a frame and attach it to the list of
        scale frames"""
        frm = Frame(self.scale_frame)
        frm.pack()
        e_pct = Entry(frm, validate="focusout",
                      validatecommand=lambda: self.validate_percentage(e_pct))
        e_pct.pack(side="left")
        e_pct.insert(0, str(percentage * 100))
        e_hex = EntryOG(frm)
        e_hex.bind(_key_release, partial(self.validate_hex_color, e_hex))
        e_hex.insert(0, hex_color)
        e_hex.configure(foreground=hex_color)
        btn_color_picker = Button(frm, text="Pick Color",
                                  command=lambda: self.color_picker_wrapper(e_hex))
        btn_color_picker.pack(side="right")
        e_hex.pack(side="right")

        if hex_color == "#FFFFFF":
            e_hex.configure(background=self.black)
        self.thresholds.append(frm)

    def color_picker_wrapper(self, e_hex: EntryOG) -> None:
        """Launches color picker and inserts into entry"""
        current_color = e_hex.get()
        new_color = askcolor(color=current_color)
        hex_code = new_color[1]
        if hex_code and isinstance(hex_code, str):
            e_hex.delete(0, END)
            e_hex.insert(0, hex_code.upper())
            self.validate_hex_color(e_hex)

    def rollback_percentage(self, e_pct: Entry, error: str, def_value: str = None) -> False:
        """Clears percentage entries"""
        e_pct.delete(0, END)
        if def_value:
            e_pct.insert(0, def_value)
        self.status.configure(text=error)
        return False

    def validate_percentage(self, e_pct: Entry) -> bool:
        """Validates percentage entries"""
        try:
            percentage = float(e_pct.get())
        except ValueError:
            return self.rollback_percentage(e_pct, "Percentage must be in range 0 - 100")

        if percentage < 0:
            return self.rollback_percentage(e_pct, "Percentage must be greater than 0",
                                            def_value="25")

        if percentage > 100:
            return self.rollback_percentage(e_pct, "Percentage must be less than 100",
                                            def_value="75")

        self.status.configure(text="")
        return False  # always validates

    def validate_hex_color(self, e_hex: EntryOG, _event: object = None) -> False:
        """Validates hex color"""
        color = e_hex.get()
        if not match("#[a-fA-F0-9]{6}$", color):
            # e_hex.delete(0, END)
            self.status.configure(text="Color code must match hex format")
            e_hex.configure(foreground=self.black, background=self.white)
        else:
            self.status.configure(text="")
    
            # calculate background based on threshold
            red, green, blue = hex_to_rgb(color)
            rgb_cutoff = (red*0.299 + green*0.587 + blue*0.114)
            e_hex.configure(foreground=color,
                            background=self.black if rgb_cutoff > 186 else self.white)

        return False  # always validates

    def increase(self):
        """Adds a new entry pair at the end with default values (100%, black)"""
        self.add_threshold(1, self.black)

    def decrease(self):
        """Removes an entry pair from the end if at least 2 pairs remain"""
        # only destroy as long as 2 pairs remain
        if len(self.thresholds) > 2:
            self.thresholds[-1].destroy()
            self.thresholds = self.thresholds[:-1]

        # set last percentage to 100 if pairs are reduced to 2
        if len(self.thresholds) == 2:
            pct = self.thresholds[-1].winfo_children()[0]
            pct.delete(0, END)
            pct.insert(0, "100")

    def set(self):
        """Validate all entries have values, reformat and set scale value, destroys popup"""
        last_child_percentage = 0
        percentage_dupe_check = []
        last_index = len(self.thresholds)

        for sf_idx, scale_frame in enumerate(self.thresholds):

            # validate all entries have values
            for idx, widget in enumerate(scale_frame.winfo_children()):
                if isinstance(widget, Entry) and widget.get() == "":
                    self.status.configure(text="All entries require valid values")
                    return

                # validate percentages are increasing
                if idx == 0:
                    this_percentage = float(widget.get())
                    if this_percentage < last_child_percentage:
                        self.status.configure(text="Threshold percentages must increase")
                        return

                    last_child_percentage = float(widget.get())

                    # validate percentages are unique
                    if this_percentage in percentage_dupe_check:
                        self.status.configure(text="Threshold percentages must be unique")
                        return

                    percentage_dupe_check.append(this_percentage)

                    # validate first percentage is 0
                    if sf_idx == 0 and this_percentage != 0.0:
                        self.status.configure(text="First percentage must be 0")
                        return

                    # validate last percentage is 100
                    if sf_idx == last_index - 1 and this_percentage != 100.0:
                        self.status.configure(text="Last percentage must be 100")
                        return

        tmp_scale = []
        for scale_frame in self.thresholds:
            pct, hex_entry, _ = scale_frame.winfo_children()
            tmp_scale.append([float(pct.get())/100, hex_entry.get()])

        # set scale in parent, recreate tooltip
        self.parent.color_scale_var.set(json.dumps(tmp_scale))
        create_tooltip(self.parent.color_scale,
                       self.parent.color_scale_tt_template + self.parent.color_scale_var.get())

        # destroy popup
        self.destroy()


class BorderPopup(Toplevel):
    """Popup to define Border properties"""
    def __init__(self, parent: [MeSHSunburst, ATCSunburst]):
        """Border Popup init"""
        super().__init__(parent)
        self.title("Set Border Properties")
        self.parent = parent
        self.resizable(False, False)
        self.error = False
        root = Frame(self)
        root.pack(padx=10, pady=10)

        # informative label
        Label(root, text="Change or disable the current border properties.").pack(pady=(10, 10))

        # get parents colors
        red, green, blue, opacity = self.parent.border_color.get().\
            replace(" ", "").lstrip("rgba(").rstrip(")").split(",")
        self.hex_color = rgb_to_hex((int(red), int(green), int(blue)))

        # colors labelframe
        col_frm = LabelFrame(root, text="Color")
        col_frm.pack(fill="x", expand=True, ipadx=2, ipady=2)
        btn_frm = Frame(col_frm)
        btn_frm.pack()
        rgb_frm = Frame(col_frm)
        rgb_frm.pack(side="left", ipadx=2, ipady=2)

        pick_color_button = Button(btn_frm, text="Pick Color", command=self.color_picker_wrapper)
        pick_color_button.pack()

        # red
        red_frm = Frame(rgb_frm)
        red_frm.pack(fill="x", expand=True, pady=(5, 2))
        Label(red_frm, text="Red").pack(side="left")
        self.red = Entry(red_frm)
        self.red.insert(0, red)
        self.red.pack(side="right")
        self.red.bind(_key_release, partial(self.validate_color, self.red))

        # green
        green_frm = Frame(rgb_frm)
        green_frm.pack(fill="x", expand=True, pady=2)
        Label(green_frm, text="Green").pack(side="left")
        self.green = Entry(green_frm)
        self.green.insert(0, green)
        self.green.pack(side="right")
        self.green.bind(_key_release, partial(self.validate_color, self.green))

        # blue
        blue_frm = Frame(rgb_frm)
        blue_frm.pack(fill="x", expand=True, pady=2)
        Label(blue_frm, text="Blue").pack(side="left")
        self.blue = Entry(blue_frm)
        self.blue.insert(0, blue)
        self.blue.pack(side="right")
        self.blue.bind(_key_release, partial(self.validate_color, self.blue))

        # rgb-hex
        hex_frm = LabelFrame(col_frm, text="Hex color")
        hex_frm.pack(side="right", ipadx=2, ipady=2, padx=(10, 4))
        self.hex = Entry(hex_frm, validate="focusout", validatecommand=self.validate_hex_color)
        self.hex.pack()
        self.preview = LabelOG(hex_frm, text="Preview")
        self.preview.pack()
        self.set_hex_from_rgb()

        # opacity
        opacity_width_frm = Frame(root)
        opacity_width_frm.pack()
        opacity_frm = Frame(opacity_width_frm)
        opacity_frm.pack(pady=(10, 2))
        Label(opacity_frm, text="Opacity [%]", width=12).pack(side="left", padx=(0, 10))
        self.opacity_var = IntVar()
        opacity_scale = ttk.LabeledScale(opacity_frm, variable=self.opacity_var, from_=0, to=100)
        opacity_scale.scale.set(int(float(opacity)*100))
        opacity_scale.pack(side="right")

        # width
        width_frm = Frame(opacity_width_frm)
        width_frm.pack(pady=2)
        Label(width_frm, text="Width [px]", width=12).pack(side="left", padx=(0, 10))
        self.width = Entry(width_frm, validate="focusout", validatecommand=self.validate_width)
        self.width.insert(0, self.parent.border_width.get())
        self.width.pack(side="right")

        self.status = Label(root, text="", foreground="red")
        self.status.pack(pady=(10, 0))

        # bottom button bar
        button_frm = Frame(root)
        button_frm.pack(fill="x", expand=True, pady=(20, 0))
        Button(button_frm, text="Cancel", command=lambda: self.destroy()).pack(side="right")
        Button(button_frm, text="Disable Border", command=self.disable).pack(side="right")
        Button(button_frm, text="Apply", command=self.set).pack(side="right")

        # freeze mainloop
        self.wait_window(self)

    def color_picker_wrapper(self) -> None:
        """Launches color picker, inserts into hex entry, triggers validation"""
        rgb, hex_color = askcolor(color=self.hex_color)
        if rgb and hex_color:
            hex_color = hex_color.upper()
            self.hex.delete(0, END)
            self.hex.insert(0, hex_color)
            self.validate_hex_color()

    def validate_hex_color(self) -> False:
        """Validates hex color"""
        if not match("#[a-fA-F0-9]{6}$", self.hex.get()):
            self.hex.delete(0, END)
            self.status.configure(text="Color code must match hex format")
        else:
            self.status.configure(text="")

            # set rgb from hex
            red, green, blue = hex_to_rgb(self.hex.get())
            self.red.delete(0, END)
            self.red.insert(0, str(red))
            self.green.delete(0, END)
            self.green.insert(0, str(green))
            self.blue.delete(0, END)
            self.blue.insert(0, str(blue))
    
            # update preview
            self.preview.configure(foreground=self.hex.get())

        return False

    def set_hex_from_rgb(self):
        """Writes hex color to entry if RGB is valid and all rgb values are within accepted range"""
        red, green, blue = self.red.get(), self.green.get(), self.blue.get()
        for color in [red, green, blue]:
            try:
                color = int(color)
                assert 0 <= color <= 255
            except (ValueError, AssertionError):
                self.error = True
                return

        self.hex.delete(0, END)
        self.hex.insert(0, rgb_to_hex((int(red), int(green), int(blue))))
        self.preview.configure(foreground=self.hex.get())

    def validate_color(self, wdg, _event) -> False:
        """Validates RGB color"""
        try:
            entry_value = wdg.get()
            color = int(entry_value)
            assert 0 <= color <= 255

            # remove preceding zeroes
            if entry_value[0] == "0" and len(entry_value) > 1:
                wdg.delete(0, END)
                wdg.insert(0, entry_value[1:])
            
            self.status.configure(text="")
            self.set_hex_from_rgb()
        except (ValueError, AssertionError):
            wdg.delete(0, END)
            wdg.insert(0, "0")
            self.status.configure(text="Colors must be decimals in range 0-255")
            self.set_hex_from_rgb()
        
        return False

    def validate_width(self) -> False:
        """Validates border width"""
        try:
            width = float(self.width.get())
            assert width >= 0
            self.status.configure(text="")
            self.error = False
        except (ValueError, AssertionError):
            self.status.configure(text="Width must be a float >= 0")
            self.error = True

        return False

    def disable(self):
        """Popup kill event"""
        self.parent.show_border_var.set(False)
        create_tooltip(self.parent.show_border, self.parent.show_border_tt_template)
        self.destroy()

    def set(self):
        """Hand over variables to parent, keeps Popup alive if validation fails"""
        if self.status["text"] != "" or self.error:
            return

        border_color = str(f"rgba({self.red.get()},{self.green.get()},"
                           f"{self.blue.get()},{float(self.opacity_var.get())/100})")
        self.parent.border_color.set(border_color)
        self.parent.border_width.set(self.width.get())
        self.parent.show_border_var.set(True)
        create_tooltip(self.parent.show_border,
                       self.parent.show_border_tt_template + "\nCurrent properties: Color: "
                       + self.parent.border_color.get()
                       + ", Width: " + self.parent.border_width.get())
        self.destroy()


class SelectOptionsPopup(Toplevel):
    """Popup to define the type of the ontology in case automatic parsing was not successful"""
    def __init__(self, parent: [MeSHSunburst, ATCSunburst] = None, title: str = None,
                 info_text: str = None, options: dict = None, is_ontology_popup: bool = False):
        super().__init__(parent)
        self.title(title)
        self.parent = parent
        self.resizable(False, False)
        self.result = None
        self.description = None
        self.is_ontology_popup = is_ontology_popup
        self.separator = None

        # custom .obo definitions
        self.custom_url = None
        self.min_node_size = None
        self.root_id = None
        self.url_error = "Enter URL to .obo file!"

        lbl_frame = Frame(self)
        lbl_frame.pack()
        descriptive_label = ttk.Label(lbl_frame, wraplength=400, text=info_text)
        descriptive_label.pack(pady=10, padx=10)

        self.radio_var = StringVar()

        self.options = options
        rb_frame = Frame(self)
        rb_frame.pack()
        for ontology_id, texts in self.options.items():
            title, tooltip = texts
            rb = Radiobutton(rb_frame, text=title, variable=self.radio_var, value=ontology_id)
            rb.pack(anchor="w")
            if tooltip:
                create_tooltip(rb, tooltip)

        sep_frame = Frame(self)
        sep_frame.pack(pady=(10, 0))
        self.sep_var = BooleanVar()
        self.sep_check = Checkbutton(sep_frame, text="Float-Counts", variable=self.sep_var, command=self.sep_controller)
        self.sep_check.pack(side="left")
        create_tooltip(self.sep_check, "Check if your counts are positive floating point values")
        self.sep_entry = Entry(sep_frame, state="disabled")
        self.sep_entry.pack(side="left")
        create_tooltip(self.sep_entry, "Enter floating point separator (. or ,)")

        self.status = Label(self, text="")
        self.status.pack()

        if self.is_ontology_popup:
            self.radio_var.trace_add("write", self.radio_var_callback)
            self.cpane = CollapsiblePane(self)
            self.cpane.pack()
            Label(self.cpane.frame, text="URL:").grid(column=0, row=0, sticky="E", padx=(0, 2))
            self.url_entry = Entry(self.cpane.frame)
            self.url_entry.grid(column=1, row=0, sticky="W")
            create_tooltip(self.url_entry, "Enter URL to .obo file, "
                                           "e.g.: https://purl.obolibrary.org/obo/po.obo"
                                           "\nCheck https://obofoundry.org/ for ontologies")
            Label(self.cpane.frame, text="Root ID:").grid(column=0, row=1, sticky="E", padx=(0, 2))
            self.root_id_entry = Entry(self.cpane.frame)
            self.root_id_entry.grid(column=1, row=1, sticky="W")
            create_tooltip(self.root_id_entry,
                           "Optional: Enter the root ID to start building a tree structure.\n"
                           "All the children of the specified ID will be used as sub-trees.\n"
                           "If the root ID field is left empty, the sub-trees will be\n"
                           "constructed based on nodes that do not have an 'is_a' relationship.")
            Label(self.cpane.frame, text="Min. Node Size:").grid(column=0, row=2, sticky="E",
                                                                 padx=(0, 2))
            self.min_node_size_entry = Entry(self.cpane.frame)
            self.min_node_size_entry.grid(column=1, row=2, sticky="W")
            create_tooltip(self.min_node_size_entry, "Optional: Enter the minimum amount of nodes "
                                                     "for a sub-tree to be included in the "
                                                     "visualization")

        btn_frame = Frame(self)
        btn_frame.pack(pady=10)

        ok_button = Button(btn_frame, text="OK", command=self.on_ok)
        ok_button.pack(side="left")

        cancel_button = Button(btn_frame, text="Cancel", command=self.on_cancel)
        cancel_button.pack(side="left")

        # freeze mainloop
        self.wait_window(self)

    def sep_controller(self):
        if self.sep_var.get():
            self.sep_entry.config(state="normal")
        else:
            self.sep_entry.config(state="disabled")
            self.sep_entry.delete(0, "end")

    def radio_var_callback(self, *args):
        if self.radio_var.get() == "custom_url":
            self.cpane.show()
        else:
            self.cpane.hide()

    def on_ok(self):
        self.status.config(text="")
        if not self.verify_result() or not self.verify_ontology_params() or not self.verify_separator_params():
            return False

        self.destroy()

    def verify_result(self):
        if not self.radio_var.get():
            self.status.config(text="Select an ontology type!")
            return False
        self.result = self.radio_var.get()
        self.description = self.options[self.result][0]
        return True

    def verify_ontology_params(self) -> bool:
        if self.radio_var.get() == "custom_url" and self.is_ontology_popup:
            url_entry = self.url_entry.get()
            min_node_size = self.min_node_size_entry.get()
            root_id = self.root_id_entry.get()

            # check URL is entered
            if not url_entry or url_entry == self.url_error or not url_entry.endswith(".obo"):
                self.url_entry.delete(0, END)
                self.url_entry.insert(0, self.url_error)
                return False

            self.custom_url = url_entry

            # ensure min_node_size is an int
            if min_node_size:
                try:
                    self.min_node_size = int(min_node_size)
                except ValueError:
                    self.min_node_size_entry.delete(0, END)
                    self.min_node_size_entry.insert(0, "Must be an integer!")
                    return False

            if root_id:
                self.root_id = root_id
        return True

    def verify_separator_params(self):
        if self.sep_var.get():
            if self.sep_entry.get() not in [",", "."]:
                self.sep_entry.delete(0, "end")
                self.status.config(text="Float separator must be defined (point . or comma ,)")
                return False

            self.separator = self.sep_entry.get() if self.sep_entry.get() else None
        return True

    def on_cancel(self):
        self.destroy()


class CollapsiblePane(ttk.Frame):
    """
     -----USAGE-----
    collapsiblePane = CollapsiblePane(parent,
                          expanded_text =[string],
                          collapsed_text =[string])

    collapsiblePane.pack()
    button = Button(collapsiblePane.frame).pack()
    """

    def __init__(self, parent):

        ttk.Frame.__init__(self, parent)
        self.parent = parent

        # Here weight implies that it can grow its size if extra space is available
        self.columnconfigure(1, weight=1)

        # dummy label, required for resizing to work properly
        self._dummy = Label(self)
        self._dummy.grid(row=0, column=0)

        self.frame = ttk.Frame(self)

    def show(self):
        self.frame.grid(row=1, column=0, columnspan=2)

    def hide(self):
        self.frame.grid_forget()



from time import sleep
from threading import Thread

import tkinter
import plotly

from src.ontoloviz.core_utils import rgb_to_hex, chunks, generate_color_range
from src.ontoloviz.core import SunburstBase, MeSHSunburst, ATCSunburst
from src.ontoloviz.app import App, BorderPopup, ExportPopup, ColorScalePopup


def test_rgb_to_hex():
    """Test utility function rgb_to_hex"""
    assert rgb_to_hex(rgb=(255, 255, 255)) == "#FFFFFF"


def test_rgb_to_hex_negative_rgb_value():
    """Test utility function rgb_to_hex when error-prone array is passed"""
    color_array = plotly.colors.n_colors(lowcolor=(64, 60, 83), highcolor=(255, 0, 255),
                                         n_colors=11001)
    negative_rgb_tuple = color_array[-1]
    assert rgb_to_hex(rgb=negative_rgb_tuple) == "#FE00FF"


def test_generate_color_range():
    """Test utility function generate_color_scale"""
    assert generate_color_range(
        start_color="#000000", stop_color="#FF0000", values=3) == ["#000000", "#7F0000", "#FF0000"]


def test_chunking():
    """Test utility function chunks"""
    assert len(list(chunks(input_list=list(range(505)), number_of_chunks=10))) == 10


def test_sunburst_class_inits():
    """Test SunburstBase and child classes"""
    assert isinstance(SunburstBase(), SunburstBase)
    assert isinstance(MeSHSunburst(), MeSHSunburst)
    assert isinstance(ATCSunburst(), ATCSunburst)


def test_drug_sunburst_attributes():
    """Test ATCSunburst class"""
    drug_sunburst = ATCSunburst()
    assert hasattr(drug_sunburst, "atc_tree")
    assert not drug_sunburst.is_init
    drug_sunburst.init()
    assert drug_sunburst.is_init


def auto_close_ui_thread(app_instance: App):
    """Thread to stop tkinter mainloop after 3 seconds"""
    sleep_timer = 0
    while sleep_timer < 3:
        sleep(1)
        print(f"Sleeping for {3 - sleep_timer} s before destroying app")
        sleep_timer += 1
    app_instance.destroy()
    app_instance.quit()


def test_ui():
    """Test visual components"""
    try:
        app = App()

        app_auto_close = Thread(target=auto_close_ui_thread, args=[app], daemon=True)
        app_auto_close.start()
        Thread(target=BorderPopup, args=[app], daemon=True).start()
        Thread(target=ExportPopup, args=[app], daemon=True).start()
        Thread(target=ColorScalePopup, args=[app], daemon=True).start()
        app.mainloop()
    except tkinter.TclError:
        print("No Window provider available - unable to test UI")
        assert True



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ontoloviz']

package_data = \
{'': ['*']}

install_requires = \
['openpyxl>=3.1.2,<4.0.0',
 'packaging>=23.0,<24.0',
 'plotly>=5.15.0,<6.0.0',
 'requests>=2.31.0,<3.0.0']

entry_points = \
{'console_scripts': ['ontoloviz = ontoloviz.app:run_app']}

setup_kwargs = {
    'name': 'ontoloviz',
    'version': '1.7.5',
    'description': 'OntoloViz drug- and phenotype-ontology visualization GUI',
    'long_description': '[![Pypi version](https://img.shields.io/pypi/v/OntoloViz.svg)](https://pypi.python.org/pypi/ontoloviz)\n[![Pypi python version](https://img.shields.io/pypi/pyversions/ontoloviz)](https://img.shields.io/pypi/pyversions/ontoloviz)\n[![Python package](https://github.com/Delta4AI/OntoloViz/actions/workflows/python-package.yml/badge.svg)](https://github.com/Delta4AI/OntoloViz/actions/workflows/python-package.yml)\n[![pylint-badge](https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/MNikley/303283c4b9026d59cda9e9dd9f697110/raw/pylint_badge.json)](https://github.com/Delta4AI/OntoloViz/actions/workflows/pylint.yml)\n<!--- documentation for pylint-badge: https://github.com/marketplace/actions/dynamic-badges --->\n\n---\n\n<div style="text-align:center">\n  <img src="https://github.com/Delta4AI/OntoloViz/assets/75040444/0b862f62-d0a6-4616-8021-2af2d0302ec2">\n  <p>\n    OntoloViz is a user-friendly interface that enables the creation of interactive sunburst plots for \n    biomedical ontologies. It allows you to conveniently visualize your data for reports or share the \n    generated plots with collaborators. Check out the <a href="#templates-and-examples">examples section</a>\n    to download interactive .html examples and to gain a better understanding of the package\'s capabilities.\n  </p>\n</div>\n\n---\nQuickstart\n==========\nThe GUI can be run by downloading the latest [release](https://github.com/Delta4AI/OntoloViz/releases) \nor by installing the package via PyPi (OS independent, requires **Python 3.8+**):\n\n    pip install ontoloviz\n\nAfter the installation you can run the GUI from the command line with the following command:\n\n    ontoloviz\n\n---\nUsage\n=====\n\nThe application allows importing `.tsv` and `.xlsx` files, but the use of `.tsv` and `tab` as a \nseparator is recommended. The GUI can create two types of sunburst diagrams to represent either phenotype or drug \nontologies, which is determined by the structure of the loaded files. Any numbers entered in the input file will be converted \nto integers and decimal points will be rounded.\n\nGUI Options\n-----------\n<img height=250px alt="gui_small" src="https://user-images.githubusercontent.com/75040444/228182954-fb48a953-ec56-46db-81ad-816d9f356206.png">\n\n**Load File**: load an `.tsv` or `.xlsx` file containing ontology data based below defined file formats\n\n**Load from Web**: load an `.obo` ontology from a pre-defined list, or define a URL to any ontology\n\n### General options\n- **Set Color Scale**: define a custom color scale for the sunburst color scaling when color propagation is active\n- **Set Border**: configures the border properties drawn around sunburst wedges\n- **Drop empty nodes** (phenotype sunburst only): drops nodes who have no further children and 0 counts\n- **Wedge Width** (drug sunburst only): switch from full outer circle (total) to count-based wedge widths (remainder)\n- **Display Labels**: controls display of labels inside sunburst wedges, available options:\n  - `all`\n  - `propagation`\n  - `drugs` (drug sunburst only)\n  - `none`\n- **Legend**: displays a weighted color bar (disabled for summary plots with specific color propagation enabled)\n\n### Propagation options\n- **Enable**: enables count- and color propagation from child to parent nodes\n- **Color**: controls color propagation by the options:\n  - `off`: color scale is based on \'Color\' column from imported file\n  - `specific`: color scale is based on the maximum values of the corresponding subtree\n  - `global`: color scale is based on the maximum values of the entire tree ontology\n  - `phenotype` (phenotype sunburst only): Only the most outer phenotype in a branch is colored\n- **Counts**: controls count propagation by the options:\n  - `off`: no counts are propagated, counts equal imported values\n  - `level`: counts are propagated up to defined level, values above threshold remain unchanged\n  - `all`: counts are propagated up to central node, imported values are corrected and overwritten\n- **Level**: controls color- and count-propagation from outer to inner levels up to defined level\n  - affects color propagation when **Color** is set to `specific` or `global`\n  - affects count propagation when **Counts** is set to `level`\n  - drug sunburst: 1 corresponds to the central node, 5 to the outermost node (=drug)\n  - phenotype sunburst: 0 corresponds to the central node, 13 to the outermost node\n\n### Summary options\n- **Enable**: displays all available subtrees in a single view\n  (resource intensive, set Labels to `none` for faster loading)\n- **Columns**: defines the amount of columns when summary plot is enabled\n\n### Save and Plot options\n- **Save**: when enabled, an interactive `.html` file containing the plot as well as a `.tsv` file containing the ontology data based on the current settings is generated for later use\n- **Plot**: Process and generate plot, opens in a Browser window\n\n---\n\nMeSH Ontology\n--------------\nThis separator-based ontology follows the principles of the [MeSH tree](https://www.nlm.nih.gov/mesh/intro_trees.html).\n- A Tree ID is defined by a **combination of three numbers or letters**, for example `C01`.\n- Levels are separated by a **dot `.`**, for example `C01.001`.\n- Ontologies **up to thirteen hierarchical levels** are supported.\n- A single phenotype end-node can be assigned to multiple parent-nodes by specifying the parents tree ids as \n  a pipe separated string in the column `Tree ID`, for example `C01.001.001|C02.001.001`.\n- If a child element is defined without a valid parent node existing, the GUI generates all parent elements with the default color and value 0. For example, the node `123.001` is automatically generated if only the child node `123.001.001` was defined. This works only if at least the most central node `123` was defined manually.\n- Counts entered in the file will be converted to integers. If a node should be displayed without counts, use `0`.\n- The loaded file must contain **7 columns** and follow the below structure to be correctly recognized:\n\n| Column Index | Header Text   | Description                                                                    |\n|--------------|---------------|--------------------------------------------------------------------------------|\n| 0            | MeSH ID       | Required primary identifier of a node in format `C01.001`                      |\n| 1            | Tree ID       | Required `\\|` pipe delimited list of Tree IDs of a node (allows 1:N mappings)  |\n| 2            | Name          | Optional label to be displayed inside the sunburst wedges                      |\n| 3            | Description   | Optional description displayed in the sunburst wedge tooltip                   |\n| 4            | Comment       | Optional comment displayed in the sunburst wedge tooltip                       |\n| 5            | Counts [Name] | Required count for wedge weights, `Name` will be used as figure title          | \n| 6            | Color         | Optional color for the sunburst wedges, must be hex-string in format `#FFFFFF` |\n\n---\n\nATC Ontology\n------------\nThis kind of sunbursts have a fixed hierarchy of 5 levels and are based on the [ATC tree](https://www.who.int/tools/atc-ddd-toolkit/atc-classification).\n- ATC codes are divided into five levels, which must follow the following naming conventions:\n  - 1st level: letter\n  - 2nd level: two numbers\n  - 3rd level: letter\n  - 4th level: letter\n  - 5th level: two numbers\n- Example ATC code: **A10BA02**\n- The hierarchy does only allow 1:1 child-parent relationships, contrary to the other ontologies. \n  For example, if the drug `deltatonin` should be assigned to the \n  parent nodes `A01AA` and `B01BB`, it must be defined twice with the ids `A01AA01` and `B01BB01`.\n- The loaded file must contain **6 columns** and follow the below structure to be correctly \n  recognized as a ATC ontology:\n\n| Column Index | Header Text   | Description                                                                    |\n|--------------|---------------|--------------------------------------------------------------------------------|\n| 0            | ATC code      | Required primary identifier of a node in format `A10BA02`                      |\n| 1            | Level         | Optional level as number, not used for building tree                           |\n| 2            | Label         | Optional label to be displayed inside the sunburst wedges                      |\n| 3            | Comment       | Optional comment displayed in the sunburst wedge tooltip                       |\n| 4            | Counts [Name] | Required count for wedge weights, `Name` will be used as figure title          |\n| 5            | Color         | Optional color for the sunburst wedges, must be hex-string in format `#FFFFFF` |\n\n---\n\nCustom Separator-based Ontologies\n---------------------------------\nOntoloViz supports loading of custom ontologies - if no known format is detected, a prompt will ask\nwhether the loaded file is a separator-based ontology or if it does contain identifiers with child- \nand parent-ids. For separator-based ontologies, the following separators are supported: `.` (dot), `,` (colon), `_` (underscore), `/` (slash).\nTo generate such an ontology, the following 5-column layout is required:\n\n| Column Index | Header Text | Description                                                                                         |\n|--------------|-------------|-----------------------------------------------------------------------------------------------------|\n| 0            | ID          | Required node identifier of a node in format `A.1` - multiple IDs can be separated with `\\|` (pipe) |\n| 1            | Label       | Optional label to be displayed inside the sunburst wedges                                           |\n| 2            | Comment     | Optional comment displayed in the sunburst wedge tooltip                                            |\n| 3            | Count       | Optional count for wedge weights                                                                    |\n| 4            | Color       | Optional color for the sunburst wedges, must be hex-string in format `#FFFFFF`                      |\n\n---\n\nCustom Parent-based Ontologies\n------------------------------\nFor loading any ontology with arbitrary IDs that do not follow a structured schema, the definition of a child-parent relationship using a 6-column layout is required:\n\n| Column Index | Header Text | Description                                                                                  |\n|--------------|-------------|----------------------------------------------------------------------------------------------|\n| 0            | ID          | Required node identifier in any format - multiple IDs can be separated with `\\|` (pipe)      |\n| 1            | Parent      | Required parent identifier in any format - if parent ID does not exist, node will be removed |\n| 2            | Label       | Optional label to be displayed inside the sunburst wedges                                    |\n| 3            | Comment     | Optional comment displayed in the sunburst wedge tooltip                                     |\n| 4            | Count       | Optional count for wedge weights                                                             |\n| 5            | Color       | Optional color for the sunburst wedges, must be hex-string in format `#FFFFFF`               |\n\n---\nTemplates and Examples\n======================\n\n| Filename                                                                                                                                                     | Description                                                                                                                                                                                                                                                       |\n|--------------------------------------------------------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|\n| [atc_example_covid_drugs_experimental.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/atc_example_covid_drugs_experimental.tsv)   | ATC-based example with data from the [DrugBank](https://go.drugbank.com) indicating experimental drugs related to COVID-19.                                                                                                                                       |\n| [atc_example_covid_drugs_trial_summary.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/atc_example_covid_drugs_trial_summary.tsv) | ATC-based example with data from [publicly available clinical trial data](https://clinicaltrials.gov/) indicating drugs tested in clinical studies, one count represents usage in one study.                                                                      |\n| [atc_template.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/atc_template.tsv)                                         | ATC-based empty template of the [ATC](https://www.who.int/tools/atc-ddd-toolkit/atc-classification) tree based on the manually curated chemical database of bioactive molecules [ChEMBL v29](https://chembl.gitbook.io/chembl-interface-documentation/downloads). |\n| [custom_template_parent_based.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/custom_template_parent_based.tsv)                   | Template for creating your own ontology based on any child and parent terms.                                                                                                                                                                                      |\n| [custom_template_separator_based.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/custom_template_separator_based.tsv)             | Template to create your own ontology based on a separator-based tree structure - for this template the underscore character `_` has been used.                                                                                                                    |\n| [mesh_example_pubmed_mapped.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/mesh_example_pubmed_mapped.tsv)                       | MeSH-based example with data from the publicly available [PubMed](https://pubmed.ncbi.nlm.nih.gov/) database of publications (title + abstract), where disease-related MeSH terms were extracted and mapped to the MeSH-tree.                                     |\n| [mesh_template.tsv](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/mesh_template.tsv)                                                 | MeSH-based empty template of the [MeSH](https://meshb.nlm.nih.gov/treeView) tree `C` and `F03`. Terms are unique and mapped to all related parent nodes.                                                                                                          |\n| [atc_example.html](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/atc_example.html)                                                   | ATC-based sample plot generated with the provided `covid_drugs_trial_summary.tsv` file.                                                                                                                                                                           |\n| [mesh_example.html](https://raw.githubusercontent.com/Delta4AI/OntoloViz/master/templates/mesh_example.html)                                                 | MeSH-based sample plot generated with the provided `covid_drugs_trial_summary.tsv` file.                                                                                                                                                                          |\n\n---\n\nScreenshots and Demos\n=====================\n<img width="90%" alt="demo_creation_of_template" src="https://user-images.githubusercontent.com/75040444/228224565-af02a994-00c7-4572-b1da-f1eeec8b6f8f.gif">\n\n> **_Demo 1:_** Minimal example for creating a phenotype based ontology\n\n<img width="90%" alt="demo_creation_of_template" src="https://user-images.githubusercontent.com/75040444/228226839-2ed34f87-7a7d-498a-9f16-fd940d05c18d.gif">\n\n> **_Demo 2:_** Showcase of some of the features available in OntoloViz (used template: <a href="https://github.com/Delta4AI/OntoloViz/files/11088919/test_tree.zip">test_tree.zip</a>)\n\n<img width="90%" alt="drug_single" src="https://user-images.githubusercontent.com/75040444/228172370-a042b9d1-81af-4172-8b41-4f8c9e9287b7.png">\n\n> **_Screenshot 1:_**  Drug sunburst plot with enabled labels, counts propagated up to level 3\n\n<img width="90%" alt="phenotype_overview" src="https://user-images.githubusercontent.com/75040444/228174582-82aaad41-f3d8-4152-8161-b8f9b1dfec67.png">\n\n> **_Screenshot 2:_**  Summary phenotype sunburst plot with tooltip, counts propagated up to the central node, color coded\n\n<table>\n    <tr>\n        <td>\n            <img height=300px alt="color_scale" src="https://user-images.githubusercontent.com/75040444/228183209-6a591a3c-8729-45c9-a73b-817dce9252c1.png">\n        </td>\n        <td>\n            <img height=300px alt="color_scale" src="https://user-images.githubusercontent.com/75040444/228183234-e6aecf82-64b4-4737-b5c3-95eb87b0fb59.png">\n        </td>\n    </tr>\n</table>\n\n> **_Screenshot 3 & 4:_**  Left: define automatic color scales based on defined counts with thresholds and hex color codes, Right: define border properties (width, opacity, colors) or disable them entirely\n\n---\n\nSpecial Thanks to\n=================\n\n* Paul Perco, who had the initial idea for this package and provided support throughout the entire process\n* Andreas Heinzel, for inspiration regarding architectural- and software-related topics\n* The Delta4 GmbH team for providing helpful inputs\n\n---\nReference\n=========\nMatthias Ley, Andreas Heinzel, Lucas Fillinger, Klaus Kratochwill, Paul Perco, OntoloViz: a GUI for interactive visualization of ranked disease or drug lists using the MeSH and ATC ontologies, Bioinformatics Advances, 2023; vbad113, https://doi.org/10.1093/bioadv/vbad113\n\n',
    'author': 'Matthias Ley',
    'author_email': 'matthias.ley@delta4.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Delta4AI/OntoloViz',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
